(* CESD Machine Definition for 
   Realizability and Compositional Compiler Correctness 
      for a Polymorphic Language
   N Benton & C-K Hur
   April 2010
*)

Require Import List Program FunctionalExtensionality Omega.
Require Import util.
Require Import ssreflect ssrbool eqtype.

Set Implicit Arguments.

(***********************************
   Definition of CESD machine
 ***********************************)

Inductive OperationM : Set := M_PLUS | M_MULT | M_MINUS | M_EQ | M_LE | M_GT.

Definition OpSemM (op: OperationM) : (nat -> nat -> nat) :=
  match op with
    | M_PLUS => plus
    | M_MULT => mult
    | M_MINUS => minus
    | M_EQ => fun n n' => if beq_nat n n' then 1 else 0
    | M_LE => fun n n' => if ble_nat n n' then 1 else 0
    | M_GT => fun n n' => if blt_nat n' n then 1 else 0
  end.

Inductive Instruction : Set :=
 (* General stack manipulation *)
 | Swap  : 
   Instruction
 | Dup   : 
   Instruction
 | Pop   : 
   Instruction
 | Push : 
   nat -> Instruction

 (* Env manipulation *)
 | PushE :
   Instruction
 | PopE :
   Instruction

 (* Integer operations *)
 | PushN : 
   nat -> Instruction
 | Op    : 
   OperationM -> Instruction

 (* Function operations *)
 | PushC  : 
   list Instruction -> Instruction
 | PushRC : 
   list Instruction -> Instruction
 | App    : 
   Instruction
 | AppNoDump :
   Instruction
 | Ret    : 
   Instruction

 (* Branch operations *) 
 | Sel   : 
   list Instruction -> list Instruction -> Instruction
 | SelNoDump :
   list Instruction -> list Instruction -> Instruction
 | Join  : 
   Instruction

 (* Pair operations *)
 | MkPair  : 
   Instruction
 | ProjFst : 
   Instruction
 | ProjSnd : 
   Instruction

 (* Syntactic equality test operations *)
 | EqTest  : 
   Instruction 
 | IsNum   :
   Instruction
.

Inductive MValue : Set :=
 | VN : nat -> MValue
 | VC : list MValue -> list Instruction -> MValue
 | VRC: list MValue -> list Instruction -> MValue
 | VP : MValue -> MValue -> MValue
.

Definition Code := list Instruction.

Definition MEnv := list MValue.

Definition Stack := list MValue.

Definition Dump := (list (Code * MEnv * Stack)) % type. 

Definition CESD := (Code * MEnv * Stack * Dump) % type.

Definition MComp := (Code * Stack) % type .

Definition VtoC (v:MValue) : MComp  := (nil,v::nil).

Definition CtoC (c:Code) : MComp := (c,nil).

Definition GetEnv (cesd:CESD) : MEnv := 
  match cesd with (_,e,_,_) => e end.

Definition RemoveEnv (cesd:CESD) : CESD := 
  match cesd with (c,_,s,d) => (c,nil,s,d) end.

Definition mergeEnv (e:MEnv) (cesd:CESD) : CESD := 
  match cesd with (c0,e0,s0,d0) => (c0,e++e0,s0,d0) end.

Definition mergeComp (comp:MComp) (cesd:CESD) : CESD := 
  match comp with (c,s) => 
    match cesd with (c0,e0,s0,d0) =>
      (c++c0,e0,s++s0,d0) 
    end
  end.

Notation "'#v' v" := (VtoC v) (at level 50).
Notation "'#c' c" := (CtoC c) (at level 50).
Infix "+~+" := mergeComp (right associativity, at level 51).
Infix "+-+" := mergeEnv (right associativity, at level 51).

Definition MValue_is_num (v: MValue) : bool :=
  if v is VN n then true else false.

Lemma mergeEnv_assoc: forall e1 e2 cesd, 
  (e1++e2) +-+ cesd = e1 +-+ (mergeEnv e2 cesd).
Proof.
move => e1 e2; do 3 case; move => /= c e s d; by rewrite app_ass.
Qed.

Lemma mergeEnv_prj_rm: forall cesd,
  (GetEnv cesd) +-+ (RemoveEnv cesd) = cesd.
Proof.
do 3 case; move => /= c e s d; by rewrite -app_nil_end.
Qed.

(****
   Operational Semantics for CESD
 ****)

Definition virtual_step (s: Stack) : nat :=
  match s with
    | VN n :: _ => n
    | _ => 0
  end.

Inductive step : CESD -> CESD -> Prop :=
 (* General stack manipulation *)
 | step_swap c e s d v1 v2 :
   step (Swap ::c, e, v1::v2::s, d) (c, e, v2::v1::s, d)
 | step_dup c e s d v :
   step (Dup ::c, e, v::s, d) (c, e, v::v::s, d)
 | step_pop c e s d v :
   step (Pop ::c, e, v::s, d) (c, e, s, d)
 | step_push c e s d idx v :
   nth_error e idx = value v -> step (Push idx :: c, e, s, d) (c, e, v::s, d)

 (* Env manipulation *)
 | step_pushe c e s d v :
   step (PushE ::c, e, v::s, d) (c, v::e, s, d)
 | step_pope c e s d v :
   step (PopE ::c, v::e, s, d) (c, e, s, d)

 (* Integer operations *)
 | step_pushn  c e s d n :
   step (PushN n :: c, e, s, d) (c, e, VN n :: s, d)
 | step_op c e s d op v1 v2 :
   step (Op  op::c,e, (VN v2) ::(VN v1) ::s, d) (c,e,VN (OpSemM op v1 v2)::s,d)

 (* Function operations *)
 | step_pushc c e s d body :
   step (PushC  body :: c, e, s, d) (c, e, VC e body :: s, d)
 | step_pushrc c e s d body :
   step (PushRC  body :: c, e, s, d) (c, e, VRC e body :: s, d)
 | step_app c e s d e' body v :
   step (App :: c, e, v :: VC e' body :: s, d) (body, v::e', nil, (c,e,s)::d)
 | step_rapp c e s d e' body v :
   step (App :: c, e, v :: VRC e' body :: s, d) 
        (body, v:: (VRC e' body) :: e', nil, (c,e,s)::d)
 | step_app_nodump c e s d e' body v :
   step (AppNoDump :: c, e, v :: VC e' body :: s, d) (body, v::e', nil, d)
 | step_rapp_nodump c e s d e' body v :
   step (AppNoDump :: c, e, v :: VRC e' body :: s, d) 
        (body, v:: (VRC e' body) :: e', nil, d)
 | step_ret c e s d c' e' s' v :
   step (Ret ::c, e, v::s, (c',e',s')::d) (c',e',v::s',d)

 (* Branch operations *) 
 | step_selfalse c e s d c1 c2 :
   step (Sel c1 c2::c, e, VN 0::s,d) (c2,e,s,(c,nil,nil)::d)
 | step_seltrue c e s d v c1 c2 : v <> VN 0 -> 
   step (Sel c1 c2::c, e, v::s,d) (c1,e,s,(c,nil,nil)::d)
 | step_selfalse_nodump c e s d c1 c2 :
   step (SelNoDump c1 c2::c, e, VN 0::s,d) (c2,e,s,d)
 | step_seltrue_nodump c e s d v c1 c2 : v <> VN 0 -> 
   step (SelNoDump c1 c2::c, e, v::s,d) (c1,e,s,d)
 | step_join c e s d c' e' s' :
   step (Join ::c, e, s, (c',e',s')::d) (c'++c, e, s, d)

 (* Pair operations *)
 | step_mkpair c e s d v1 v2 :
   step (MkPair ::c, e, v2::v1::s, d) (c, e, VP v1 v2::s, d)
 | step_fst c e s d v1 v2 :
   step (ProjFst ::c, e, VP v1 v2::s, d) (c, e, v1::s, d)
 | step_snd c e s d v1 v2 :
   step (ProjSnd ::c, e, VP v1 v2::s, d) (c, e, v2::s, d)

 (* Syntactic equality test operation *)
 | step_eqtrue c e s d v1 v2 : v1=v2 -> 
   step (EqTest ::c,e,v1::v2::s,d) (c,e,VN 1::s,d) 
 | step_eqfalse c e s d v1 v2 : v1<>v2 -> 
   step (EqTest ::c,e,v1::v2::s,d) (c,e,VN 0::s,d) 
 | step_isnumtrue c e s d v : MValue_is_num v -> 
   step (IsNum ::c,e,v::s,d) (c,e,VN 1::v::s,d) 
 | step_isnumfalse c e s d v : ~ MValue_is_num v -> 
   step (IsNum ::c,e,v::s,d) (c,e,VN 0::v::s,d) 
.

Inductive kstepreduce: nat -> CESD -> CESD -> Prop :=
  | kstep0 : forall cesd, kstepreduce 0 cesd cesd
  | kstepS : forall k cesd cesd1 cesd', 
    step cesd cesd1 -> kstepreduce k cesd1 cesd' -> kstepreduce (S k) cesd cesd'
.
Hint Constructors kstepreduce.

Definition reduce cesd cesd' := exists k, kstepreduce k cesd cesd'.

Definition stepsfor k cesd := exists k', exists cesd', 
  k' >= k /\ kstepreduce k' cesd cesd'.

Definition diverge cesd := forall k, stepsfor k cesd.

Definition terminate cesd := ~ (diverge cesd).

(*
   Properties of kstepreduce
*)

Lemma step_uniq: forall cesd cesd' cesd'', 
  step cesd cesd' -> step cesd cesd'' -> cesd' = cesd''.
Proof.
intros; destruct H; 
  try (depdestr H0); auto;
  try (contradiction H); try (contradiction H0); auto.
rewrite H in H0; case: H0; by move ->.
Qed.

Lemma kstepreduce_uniq: forall k cesd cesd' cesd'', 
  kstepreduce k cesd cesd' -> kstepreduce k cesd cesd'' -> (cesd' = cesd'').
Proof.
intro gk; generalize_induction gk; induction gk; intros.
- by inversion H; subst; depdestr H0, H1.
- destruct k; depdestr H0, H1; first by [].
  rewrite -> (step_uniq H0 H1) in *.
  eapply IHgk; first last; try Apply; omega.
Qed.

Lemma kstepreduce_compose: forall k1 k2 cesd cesd1 cesd2,
  kstepreduce k1 cesd cesd1 -> kstepreduce k2 cesd1 cesd2 -> 
  kstepreduce (k1+k2) cesd cesd2.
Proof.
intros; induction H; first by [].
by econstructor; Apply.
Qed.

Lemma kstepreduce_inv: forall k j cesd cesd' cesd'',
  kstepreduce k cesd cesd' -> kstepreduce (k+j) cesd cesd''-> 
  kstepreduce j cesd' cesd''.
Proof.
intros; induction H; first by [].
by apply IHkstepreduce; depdestr H0; rewrite (step_uniq H H0).
Qed.

Lemma stepsfor_zero: forall cesd, stepsfor 0 cesd.
Proof.
move => cesd; exists 0; exists cesd; split; by [auto | constructor].
Qed.

Lemma stepsfor_decr: forall cesd k k', k <= k' -> 
  stepsfor k' cesd -> stepsfor k cesd.
Proof.
move => cesd k k' Hk [n [cesd' [Hn Hred]]].
exists n; exists cesd'; split; by [omega |].
Qed.

Lemma stepsfor_inv: forall k cesd,
  stepsfor (S k) cesd -> exists cesd', step cesd cesd' /\ stepsfor k cesd'.
Proof.
intros; destruct H as [n [cesd' [? ?]]].
destruct n; first inversion H. 
depdestr H0; eexists; split; first apply H0.
do 2 eexists; split; first last; by [apply H1 | omega].
Qed.

Lemma stepsfor_step_strict: forall cesd cesd' k,
  step cesd cesd' -> stepsfor (Peano.pred k) cesd' -> stepsfor k cesd.
Proof.
intros; destruct H0 as [n [cesd'' [? ?]]].
do 2 eexists; split; first last; by [eapply kstepS; Apply | omega].
Qed.

Corollary stepsfor_step: forall cesd cesd' k,
  step cesd cesd' -> stepsfor k cesd' -> stepsfor k cesd.
Proof.
intros; eapply stepsfor_step_strict.
- apply H.
- apply stepsfor_decr with k; by [omega | auto].
Qed.

Lemma terminate_step: forall cesd cesd', 
  step cesd cesd' -> terminate cesd' -> terminate cesd.
Proof.
repeat red; intros; apply H0.
move => k; replace_hyp H1 (H1 (S k)); destruct H1 as [n [cesd'' [? ?]]].
destruct H2; first by inversion H1. 
rewrite -> (step_uniq H H2) in *.
do 2 eexists; split; first last; by [Apply | omega].
Qed.

Lemma terminate_kstepreduce: forall k cesd cesd',
  kstepreduce k cesd cesd' -> terminate cesd -> terminate cesd'.
Proof.
do 2 red; intros; apply H0.
move => k0; destruct (H1 k0) as [k' [cesd'' [? ?]]].
exists (k + k'); exists cesd''; split; first by omega.
eapply kstepreduce_compose; by [apply H |].
Qed.

Lemma stepsfor_kstepreduce: forall i j k cesd cesd',
  kstepreduce i cesd cesd' -> stepsfor j cesd -> j >= i+k ->stepsfor k cesd'.
Proof.
intros; destruct H0 as [n [cesd'' [? ?]]]. 
exists (n-i); exists cesd''; split; first by omega.
eapply kstepreduce_inv; first by apply H.
by replace (i+(n-i)) with n by omega.
Qed.

Lemma terminate_kstepreduce_rev: forall k cesd cesd',
  kstepreduce k cesd cesd' -> terminate cesd' -> terminate cesd.
Proof.
do 2 red; intros; apply H0; red; intros.
destruct (H1 (k+k0)) as [k' [cesd'' [? ?]]].
exists (k'-k); exists cesd''; split; first by omega.
eapply kstepreduce_inv; first apply H.
by replace (k+(k'-k)) with k' by omega.
Qed.

Lemma stepsfor_kstepreduce_rev: forall i j cesd cesd',
  kstepreduce i cesd cesd' -> stepsfor (j-i) cesd' -> stepsfor j cesd.
Proof.
intros; destruct H0 as [n [cesd'' [? ?]]].
exists (i+n); exists cesd''; split; first by omega.
eapply kstepreduce_compose; by [apply H | ].
Qed.

(*
   CESD execution tactic
*)

Ltac cesd_step := rewrite <- ?app_nil_end; solve
 [ constructor; simpl; auto; 
   solve [discriminate; reflexivity | try apply nth_error_prefix; Apply]
 | apply step_eqfalse; simpl; auto; 
   solve [discriminate; reflexivity | try apply nth_error_prefix; Apply]
 | apply step_isnumfalse; simpl; auto; 
   solve [discriminate; reflexivity | try apply nth_error_prefix; Apply] ].
Ltac kstep := eapply kstepS; first by cesd_step.
Ltac sstep := eapply stepsfor_step; first by cesd_step.
Ltac tstep := eapply terminate_step; first by cesd_step.
Ltac ksteps := repeat kstep.
Ltac ssteps := repeat sstep.
Ltac tsteps := repeat tstep.

Ltac mkcesd c e s d := 
  evar (c:Code); evar (e:MEnv); evar (s:Stack); evar (d:Dump).
Ltac rmcesd c e s d := (unfold c,e,s,d in *; clear c e s d).
Tactic Notation "withcesd" ident(c) ident(e) ident(s) ident(d) tactic(t) :=
  (mkcesd c e s d; t; rmcesd c e s d).
