(****
   Logical relations for
   Realizability and Compositional Compiler Correctness 
      for a Polymorphic Language
   N Benton & C-K Hur
   April 2010
 ****)

Require Import Arith Max List Program FunctionalExtensionality.
Require Import util cesd sysf sysfopsem.
Require Import ssreflect ssrbool Heq.

Set Implicit Arguments.

(* indexed predicate, relation, subset *)

Definition prdc (A:Type) := A->Prop.

Definition indexed (A:Type) := (nat * A) % type.

Definition idecr A (P:prdc (indexed A)) := 
  forall k k' a, k <= k' -> P (k',a) -> P (k,a).

Definition irel (A B:Type) := nat -> A -> B -> Prop.

Definition rel (A B:Type) := A -> B -> Prop.

Definition irelId (A: Type) : irel A A := fun i a a' => a = a'.

Definition ireldecr A B (r: irel A B) := 
  forall k k' a b, k <= k' -> r k' a b -> r k a b.

Definition subset (A : Type) (p q : A -> Prop) := forall a, p a -> q a.

Definition subset2 (A B: Type) (p q : A -> B -> Prop) := 
  forall a b, p a b -> q a b.

Definition subset3 (A B C: Type) (p q : A -> B -> C -> Prop) := 
  forall a b c, p a b c -> q a b c.

(****************************************

  Observational Approximation Signature

 ****************************************)

Section TRelDefs.

  Definition stream (A:Type) : Type := nat -> A.
  Definition cstr (A:Type) (a:A) : stream A := fun _ => a.
  Definition is_chain (A:Type) (le: rel A A) (s: stream A) : Prop := 
    forall i j, i <= j -> le (s i) (s j).

  Definition TRelV := forall u E T, Val u E T -> Val u E T -> Prop.
  Definition TRelC := forall u E T, Exp u E T -> Exp u E T -> Prop.
  Definition STRelV := forall u E T, Val u E T -> stream (Val u E T) -> Prop.
  Definition STRelC := forall u E T, Exp u E T -> stream (Exp u E T) -> Prop.

  Variable LeV : TRelV.
  Variable LeC : TRelC.

  Variable SLeV : STRelV.
  Variable SLeC : STRelC.

  Notation "tv <v< tv'" := (LeV tv tv') (at level 57).
  Notation "tv <c< tv'" := (LeC tv tv') (at level 57).
  Notation "tv =c= tv'" := (LeC tv tv' /\ LeC tv' tv) (at level 57).

  Notation "tv <<v<< tvs" := (SLeV tv tvs) (at level 57).
  Notation "tv <<c<< tvs" := (SLeC tv tvs) (at level 57).

  (* Reflexivity for LeV and LeC *)
  Definition TRelV_refl := forall u E T (tv: Val u E T), tv <v< tv.
  Definition TRelC_refl := forall u E T (tv: Exp u E T), tv <c< tv.

  (* Beta reduction for LeV and LeC *)
  Definition TRel_beta := forall u E A B tv (tc : Exp u (A::E) B),
    LET (^tv) tc =c= SubExp {|tv|} tc.

  (* Chain property for SLeV and SLeC *)
  Definition STRelV_chain := forall u E T (tv: Val u E T) tvs,
    tv <<v<< tvs -> is_chain (@LeV _ _ _) tvs.
  Definition STRelC_chain := forall u E T (tv: Exp u E T) tcs,
    tv <<c<< tcs -> is_chain (@LeC _ _ _) tcs.

  (* Reflexivity for SLeV and SLeC *)
  Definition STRelV_refl := forall u E T (tvs: stream (Val u E T)) p, 
    is_chain (@LeV _ _ _) tvs -> tvs p <<v<< tvs.
  Definition STRelC_refl := forall u E T (tvs: stream (Exp u E T)) p, 
    is_chain (@LeC _ _ _) tvs -> tvs p <<c<< tvs.

  (* Transitivity for SLeV and SLeC *)
  Definition STRelV_trans := forall u E T (tv: Val u E T) tvs tvs',
    tv <<v<< tvs -> (forall p, tvs p <<v<< tvs') -> tv <<v<< tvs'.
  Definition STRelC_trans := forall u E T (tv: Exp u E T) tvs tvs',
    tv <<c<< tvs -> (forall p, tvs p <<c<< tvs') -> tv <<c<< tvs'.

  (* Substitutivity for SLeV and SLeC *)
  Definition STRelV_sub := 
    forall u w (sub: subT u w) (E:Env u) T (tv: Val u E T) tvs,
    tv <<v<< tvs -> subVal sub tv <<v<< fun p => subVal sub (tvs p).
  Definition STRelC_sub :=
    forall u w (sub: subT u w) (E:Env u) T (tv: Exp u E T) tvs,
    tv <<c<< tvs -> subExp sub tv <<c<< fun p => subExp sub (tvs p).
  Definition STRelV_Sub :=
    forall u E E' (Sub: SubT E E') Subs T (tv: Val u E T) tvs,
    (forall T (var: VarT E T), Sub _ var <<v<< fun p => (Subs p) _ var) ->
    tv <<v<< tvs -> 
    SubVal Sub tv <<v<< fun p => SubVal (Subs p) (tvs p).
  Definition STRelC_Sub :=
    forall u E E' (Sub: SubT E E') Subs T (tv: Exp u E T) tvs,
    (forall T (var: VarT E T), Sub _ var <<v<< fun p => (Subs p) _ var) -> 
    tv <<c<< tvs -> 
    SubExp Sub tv <<c<< fun p => SubExp (Subs p) (tvs p).
  
  (* Compatibility for SLeV and SLeC *)
  Definition STRel_VAR := forall u (E: Env u) T (var: VarT E T),
    VAR var <<v<< cstr (VAR var).
  Definition STRel_UNIT := forall u (E: Env u),
    VUNIT <<v<< cstr (VUNIT (env:=E)).
  Definition STRel_INT := forall u (E: Env u) n,
    VINT n <<v<< cstr (VINT (env:=E) n).
  Definition STRel_PAIR :=
    forall u E T1 T2 (tv1: Val u E T1) tvs1 (tv2: Val u E T2) tvs2,
    tv1 <<v<< tvs1 -> tv2 <<v<< tvs2 -> 
    PAIR tv1 tv2 <<v<< fun p => PAIR (tvs1 p) (tvs2 p).
  Definition STRel_INL := forall u E T1 T2 (tv1: Val u E T1) tvs1,
    tv1 <<v<< tvs1 -> INL T2 tv1 <<v<< fun p => INL T2 (tvs1 p).
  Definition STRel_INR := forall u E T1 T2 (tv2: Val u E T2) tvs2,
    tv2 <<v<< tvs2 -> INR T1 tv2 <<v<< fun p => INR T1 (tvs2 p).
  Definition STRel_LAM := forall u (E: Env u) T1 T2 (tv: Exp u (T1::E) T2) tvs,
    tv <<c<< tvs -> LAM tv <<v<< fun p => LAM (tvs p).
  Definition STRel_REC :=
    forall u (E: Env u) T1 T2 (tv: Exp u (T1::T1-->T2::E) T2) tvs,
    tv <<c<< tvs -> REC tv <<v<< fun p => REC (tvs p).
  Definition STRel_TLAM :=
    forall u (E:Env u) (T:Ty (S u)) (tv: Val _ (subEnv shsub E) T) tvs,
    tv <<v<< tvs -> TLAM tv <<v<< fun p => TLAM (tvs p).
  Definition STRel_PACK :=
    forall u (E:Env u) (T:Ty (S u)) t (tv: Val _ E (subTy [| t |] T)) tvs,
    tv <<v<< tvs -> PACK tv <<v<< fun p => PACK (tvs p).
  Definition STRel_VAL := forall u E T (tv: Val u E T) tvs,
    tv <<v<< tvs -> ^tv <<c<< fun p => ^tvs p.
  Definition STRel_OPN := forall u E op (tv1 tv2 : Val u E INT) tvs1 tvs2,
    tv1 <<v<< tvs1 -> tv2 <<v<< tvs2 -> 
    OPN op tv1 tv2 <<c<< fun p => OPN op (tvs1 p) (tvs2 p).
  Definition STRel_OPB := forall u E op (tv1 tv2 : Val u E INT) tvs1 tvs2,
    tv1 <<v<< tvs1 -> tv2 <<v<< tvs2 -> 
    OPB op tv1 tv2 <<c<< fun p => OPB op (tvs1 p) (tvs2 p).
  Definition STRel_FST := forall u E T1 T2 (tv: Val u E (T1 *** T2)) tvs,
    tv <<v<< tvs -> FST tv <<c<< fun p => FST (tvs p).
  Definition STRel_SND := forall u E T1 T2 (tv: Val u E (T1 *** T2)) tvs,
    tv <<v<< tvs -> SND tv <<c<< fun p => SND (tvs p).
  Definition STRel_CASE := forall u E T1 T2 T (tv: Val u E (T1 +++ T2)) tvs 
    (tv1: Exp u (T1::E) T) (tv2: Exp u (T2::E) T) tvs1 tvs2,
    tv <<v<< tvs -> tv1 <<c<< tvs1 -> tv2 <<c<< tvs2 -> 
    CASE tv tv1 tv2 <<c<< fun p => CASE (tvs p) (tvs1 p) (tvs2 p).
  Definition STRel_LET :=
    forall u E T1 T2 (tv1:Exp u E T1) tvs1 (tv2: Exp u (T1::E) T2) tvs2,
    tv1 <<c<< tvs1 -> tv2 <<c<< tvs2 -> 
    LET tv1 tv2 <<c<< fun p => LET (tvs1 p) (tvs2 p).
  Definition STRel_APP := 
    forall u E T1 T2 (tv1: Val u E (T1-->T2)) tvs1 tv2 tvs2,
    tv1 <<v<< tvs1 -> tv2 <<v<< tvs2 -> 
    APP tv1 tv2 <<c<< fun p => APP (tvs1 p) (tvs2 p).
  Definition STRel_TAPP := forall u E T (tv: Val u E (ALL T)) tvs ty,
    tv <<v<< tvs -> TAPP tv ty <<c<< fun p => TAPP (tvs p) ty.
  Definition STRel_UNPK := forall u E T t (tv1: Val u E (EXIST t)) tvs1 
    (tv2: Exp _ (t::subEnv shsub E) (subTy shsub T)) tvs2,
    tv1 <<v<< tvs1 -> tv2 <<c<< tvs2 -> 
    UNPK tv1 tv2 <<c<< fun p => UNPK (tvs1 p) (tvs2 p).

  (* Adequacy for SLeC *)
  Definition STRel_adeq := forall (tv: CExp UNIT) tvs,
    tv <<c<< tvs -> tv |==> -> exists p, tvs p |==>.

  (* Unwinding for SLeV *)
  Definition STRel_unwinding := 
    forall u (E: Env u) T1 T2 (tc: Exp u (T1::T1-->T2::E) T2),
    REC tc <<v<< fun p => RECN p tc.

End TRelDefs.

Definition to_TRelV (LeV: STRelV) : TRelV := 
  fun _ _ _ tv tv' => LeV _ _ _ tv (cstr tv').
Definition to_TRelC (LeC: STRelC) : TRelC := 
  fun _ _ _ tv tv' => LeC _ _ _ tv (cstr tv').
Definition to_STRelV (LeV: TRelV) : STRelV := 
  fun _ _ _ tv tvs => exists tv', tvs = cstr tv' /\ LeV _ _ _ tv tv'.
Definition to_STRelC (LeC: TRelC) : STRelC := 
  fun _ _ _ tv tvs => exists tv', tvs = cstr tv' /\ LeC _ _ _ tv tv'.

Definition subrel_TRelV (rel rel' : TRelV) :=
  forall u E T (tv:Val u E T) tv', rel _ _ _ tv tv' -> rel' _ _ _ tv tv'.
Definition subrel_TRelC (rel rel' : TRelC) :=
  forall u E T (tv:Exp u E T) tv', rel _ _ _ tv tv' -> rel' _ _ _ tv tv'.
Definition subrel_STRelV (rel rel' : STRelV) :=
  forall u E T (tv:Val u E T) tvs, rel _ _ _ tv tvs -> rel' _ _ _ tv tvs.
Definition subrel_STRelC (rel rel' : STRelC) :=
  forall u E T (tv:Exp u E T) tvs, rel _ _ _ tv tvs -> rel' _ _ _ tv tvs.

Record ObsLe := mkObsLe
{
  ObsLeV :> STRelV;
  ObsLeC :> STRelC;

  ObsLeV_refl      : TRelV_refl (to_TRelV ObsLeV);
  ObsLeC_refl      : TRelC_refl (to_TRelC ObsLeC);

  ObsLe_beta       : TRel_beta (to_TRelC ObsLeC);

  ObsLeV_chain     : STRelV_chain (to_TRelV ObsLeV) ObsLeV;
  ObsLeC_chain     : STRelC_chain (to_TRelC ObsLeC) ObsLeC;

  ObsLeV_elmt      : STRelV_refl (to_TRelV ObsLeV) ObsLeV;
  ObsLeC_elmt      : STRelC_refl (to_TRelC ObsLeC) ObsLeC;

  ObsLeV_trans     : STRelV_trans ObsLeV;
  ObsLeC_trans     : STRelC_trans ObsLeC;

  ObsLeV_sub       : STRelV_sub ObsLeV;
  ObsLeC_sub       : STRelC_sub ObsLeC;
  ObsLeV_Sub       : STRelV_Sub ObsLeV;
  ObsLeC_Sub       : STRelC_Sub ObsLeV ObsLeC;

  ObsLe_VAR        : STRel_VAR ObsLeV;
  ObsLe_UNIT       : STRel_UNIT ObsLeV;
  ObsLe_INT        : STRel_INT ObsLeV;
  ObsLe_PAIR       : STRel_PAIR ObsLeV;
  ObsLe_INL        : STRel_INL ObsLeV;
  ObsLe_INR        : STRel_INR ObsLeV;
  ObsLe_LAM        : STRel_LAM ObsLeV ObsLeC;
  ObsLe_REC        : STRel_REC ObsLeV ObsLeC;
  ObsLe_TLAM       : STRel_TLAM ObsLeV;
  ObsLe_PACK       : STRel_PACK ObsLeV;
  ObsLe_VAL        : STRel_VAL ObsLeV ObsLeC;
  ObsLe_OPN        : STRel_OPN ObsLeV ObsLeC;
  ObsLe_OPB        : STRel_OPB ObsLeV ObsLeC;
  ObsLe_FST        : STRel_FST ObsLeV ObsLeC;
  ObsLe_SND        : STRel_SND ObsLeV ObsLeC;
  ObsLe_CASE       : STRel_CASE ObsLeV ObsLeC;
  ObsLe_LET        : STRel_LET ObsLeC;
  ObsLe_APP        : STRel_APP ObsLeV ObsLeC;
  ObsLe_TAPP       : STRel_TAPP ObsLeV ObsLeC;
  ObsLe_UNPK       : STRel_UNPK ObsLeV ObsLeC;

  ObsLe_adeq       : STRel_adeq ObsLeC;

  ObsLe_unwinding  : STRel_unwinding ObsLeV
}.

Definition ObsLeE ole u (E:Env u) : rel (EVal E) (stream (EVal E)) := 
  fun te tes => forall T var, ObsLeV ole (te T var) (fun p => (tes p) T var).

Lemma ObsLeE_refl: forall ole u (E:Env u) (te: EVal E), ObsLeE ole te (cstr te).
Proof. by red; intros; apply ObsLeV_refl. Qed.

Lemma ObsLeE_trans: 
  forall ole u (E:Env u) (te: EVal E) (tes tes': stream (EVal E)),
  ObsLeE ole te tes -> (forall p, ObsLeE ole (tes p) tes') -> ObsLeE ole te tes'.
Proof. 
  red; intros; eapply ObsLeV_trans; first by Apply. 
  by intros; unfold cstr; Apply. 
Qed.

Lemma ObsLe_RECN: 
  forall ole i u (E: Env u) T1 T2 (tc: Exp u (T1::T1-->T2::E) T2) tcs,
  ObsLeC ole tc tcs -> ObsLeV ole (RECN i tc) (fun p => RECN i (tcs p)).
Proof.
induction i; intros; simpl RECN; first by apply ObsLeV_refl.
apply (ObsLe_LAM _ _ (fun p => SubExp (SubL T1 {|RECN i (tcs p)|}) (tcs p))).
apply (ObsLeC_Sub _ _ (fun p => SubL T1 {|RECN i (tcs p)|}) _ tcs); last by [].
intros; depdestr var; first by apply ObsLeV_refl.
depdestr var; simpl; rewrite RenVal_def; last by apply ObsLeV_refl.
apply (ObsLeV_Sub _ _ (cstr (shSub T1)) (RECN i tc) (fun p => RECN i (tcs p))).
- by intros; apply ObsLeV_refl.
- by unfold cstr in IHi; apply IHi.
Qed.

Lemma RECN_monotone: 
  forall ole u (E: Env u) T1 T2 (tc tc': Exp u (T1::T1-->T2::E) T2) i j,
  i <= j -> ObsLeC ole tc (cstr tc') -> 
  ObsLeV ole (RECN i tc) (cstr (RECN j tc')).
Proof.
intros; eapply ObsLeV_trans.
- apply (ObsLeV_chain _ _ _ (ObsLe_unwinding _ tc)); apply H.
- intros; unfold cstr; by apply ObsLe_RECN.
Qed.

Lemma REC_approximation: 
  forall ole u env ty1 ty2 (tc: Exp u (ty1::ty1-->ty2::env) ty2) tcs,
  ObsLeC ole tc tcs -> ObsLeV ole (REC tc) (fun i : nat => RECN i (tcs i)).
Proof.
intros; eapply ObsLeV_trans; first by apply ObsLe_unwinding. 
intros; simpl; eapply ObsLeV_trans; first by apply (ObsLe_RECN _ _ _ _ H).
intros; eapply ObsLeV_trans.
- unfold cstr; apply RECN_monotone.
  - by instantiate (1:= max p p0); auto with arith.
  - by apply (ObsLeC_chain _ _ _ H); instantiate (1:= max p p0); auto with arith.
- intros; unfold cstr; apply (ObsLeV_elmt _ (tvs := fun p => RECN p (tcs p))).
  red; intros; apply RECN_monotone; by [ | apply (ObsLeC_chain _ _ _ H)].
Qed.

Lemma ObsLe_adeq_unit: forall ole (tc: CExp UNIT) tcs tv,
  ObsLeC ole tc tcs -> tc |==> tv -> exists p, tcs p |==> tv.
Proof.
intros; depdestr tv; first by inversion v.
have Htc : tc |==> by eexists; Apply.
destruct (ObsLe_adeq ole _ H Htc) as [p [tv ?]].
by exists p; eapply eq_ind; [Apply | clear H1; depdestr tv; [inversion v | ]].
Qed.

Lemma ObsLe_adeq_int: forall ole (tc: CExp INT) tcs tv,
  ObsLeC ole tc tcs -> tc |==> tv -> exists p, tcs p |==> tv.
Proof.
intros; depdestr tv; first by inversion v.
have Hrel : ObsLeC ole (LET tc (LET (OPB EQ (#V 0) (VINT n)) 
   (CASE (#V 0) (^VUNIT) (APP (BotF _ UNIT UNIT) VUNIT))))
  (fun p => LET (tcs p) (LET (OPB EQ (#V 0) (VINT n)) 
   (CASE (#V 0) (^VUNIT) (APP (BotF _ UNIT UNIT) VUNIT))))
  by apply ObsLe_LET; [ | apply ObsLeC_refl].
eprop.
- apply (ObsLe_adeq _ _ Hrel).
  destruct H0 as [k ?]; do 2 eexists; econstructor; first by Apply.
  by repeat (unfold OpSemBool; rewrite ?beq_nat_true; unfold_SubVE; 
    econstructor; first by constructor).
move => [p [tv [k Hterm]]].
inversion Hterm; clear Hterm; Hsubst; move : H8; unfold_SubVE; move => Hterm'.
have Hint : forall tv : CVal INT, exists n, tv = VINT n
 by move => tv0; depdestr tv0; [inversion v| eexists; auto ].
destruct (Hint v1) as [m ?]; subst; clear Hint.
have : m = n \/ m <> n
 by destruct (lt_eq_lt_dec m n) as [[? | ?] | ?]; [right|left|right]; try omega.
case; intros; first by subst; do 2 eexists; Apply.
inversion Hterm'; Hsubst; inversion H7; Hsubst; clear Hterm' H7.
rewrite beq_nat_false in H10; last by [].
move :H10; unfold_SubVE; move =>H10; inversion H10; Hsubst; clear H10.
move :H6; unfold_SubVE; move =>H6; replace_hyp H6 (Ev_to_Treduce H6).
destruct (BotF_diverge (S k) UNIT VUNIT) as [k' [tm' [? ?]]].
replace k' with (k+(k'-k)) in H3 by omega.
have Hred := Treduce_inv H6 _ H3.
by depdestr Hred; [absurd False; [auto| omega] | inversion H4].
Qed.

Module Type ObsLeSig.

  Parameter ole : ObsLe.

  Notation "tv <<V<< tvs" := (ObsLeV ole tv tvs) (at level 70, no associativity).
  Notation "te <<E<< tes" := (ObsLeE ole te tes) (at level 70, no associativity).
  Notation "tc <<C<< tcs" := (ObsLeC ole tc tcs) (at level 70, no associativity).

  Notation "tv1 ==V== tv2" := (tv1 <<V<< (cstr tv2) /\ tv2 <<V<< (cstr tv1)) 
    (at level 70, no associativity).
  Notation "te1 ==E== te2" := (te1 <<E<< (cstr te2) /\ te2 <<E<< (cstr te1)) 
    (at level 70, no associativity).
  Notation "tc1 ==C== tc2" := (tc1 <<C<< (cstr tc2) /\ tc2 <<C<< (cstr tc1)) 
    (at level 70, no associativity).

End ObsLeSig.

(***********************************

   Logical Relation Functor

 ***********************************)

Module LogicalRelationFunctor (ObsLeMod : ObsLeSig).

Export ObsLeMod.

Definition iVal := indexed MValue.
Definition iComp := indexed MComp.
Definition iCESD  := indexed CESD.

Definition SirelV (ty:Ty 0) : Type := irel MValue (CVal ty).
Definition SirelE {u} (env:Env u) : Type := irel MEnv (EVal env).

Definition SrelV (ty:Ty 0) : Type := rel MValue (CVal ty).
Definition SrelE {u} (env:Env u) : Type := rel MEnv (EVal env).

(* 
   Top Top semantics 
*)

(**
   << Relation
 **)

(* iSubRel*)

Definition iSRel (ty:Ty 0) : Type := irel MValue (CVal ty).

Definition iSubRel u := varT u -> sigT iSRel.

Definition isr_sub u (srel:iSubRel u) : subT u 0 := fun var => projT1 (srel var).

Definition isr_rel u (srel:iSubRel u) := fun var => projT2 (srel var).

Definition isr_cons u ty tr (srel: iSubRel u) : iSubRel (S u) :=
  @consmap (fun _ => sigT iSRel) u 0 ({{ iSRel # ty , tr }}) srel.
Implicit Arguments isr_cons [u].

Definition isr_decr u (srel:iSubRel u) := 
  forall var, ireldecr (isr_rel srel var).

Definition isr0 : iSubRel 0 := fun var => varT_empty _ var.

Lemma isr_sub_cons: forall u t tr (srel: iSubRel u),
  isr_sub (isr_cons t tr srel) = consmap t (isr_sub srel).
Proof. intros; mapext_destr. Qed.

Lemma isr0_idsub: isr_sub isr0 = idsub.
Proof. mapext_destr. Qed.

Lemma isr_singsub: forall u (srel:iSubRel u) ty,
  isr_sub srel @ss@ [|ty|] = consmap (subTy (isr_sub srel) ty) (isr_sub srel).
Proof. intros; mapext_destr. Qed.

(**
   biorthogonaltiy 
**)

Definition pton_kstepv (pe:MEnv) (P: prdc iVal) : prdc iCESD  := fun jcesd =>
  match jcesd with (j,(c,e,s,d)) =>
    forall i val, P (i, val) ->
      forall k, k <= i -> k <= j -> stepsfor k (c, pe++e, val::s, d)
  end.
  
Definition ntop_kstepc (pe:MEnv) (P: prdc iCESD) : prdc iComp := fun icomp =>
  match icomp with (i,(c1,s1)) =>
    forall j c e s d, P (j, (c,e,s,d)) ->
      forall k, k <= i -> k <= j -> stepsfor k (c1++c, pe++e, s1++s, d)
  end.

Definition irel_to_ipred (A B:Type) (r:irel A B) (b:B) : prdc (indexed A) := 
  fun ia => match ia with (i,a) => r i a b end.

Definition irel_lift T (RT: irel MValue (CVal T)) : irel MValue (CExp T) := 
  fun i val tc => exists tv, tc |==> tv /\ RT i val tv.

Definition ivalset_le T (RT: irel MValue (CVal T)) (exp:CExp T) : prdc iVal := 
  irel_to_ipred (irel_lift RT) exp.

Definition MonadLe (e:MEnv) T (RT: irel MValue (CVal T))
  : irel MComp (CExp T) :=
  fun i comp tc => ntop_kstepc e (pton_kstepv e (ivalset_le RT tc)) (i, comp).


(*
   Approximation relation
*)

(* irelLe Functor *)

Definition irelLeV_C1: forall u (srel: iSubRel u) T t (tr: irel MValue (CVal t)),
  (nil:Env 0, (subTy [|t|] (subTy (subL (isr_sub srel)) T))) =
  (nil:Env 0, (subTy (isr_sub (isr_cons t tr srel)) T)).
Proof. by intros; simpl_sub; rewrite -> !isr_sub_cons. Defined.
Implicit Arguments irelLeV_C1 [ u srel T t tr ].

Definition irelLeVFun_body 
  (irelLeVFun: forall u (srel: iSubRel u) (T:Ty u), 
    irel MValue (CVal (subTy (isr_sub srel) T))) 
  u (srel: iSubRel u) (T:Ty u) : irel MValue (CVal (subTy (isr_sub srel) T))
 :=
  match T return irel MValue (CVal (subTy  (isr_sub srel) T)) with
    | TYVAR var => isr_rel srel var
    | UNIT => fun _ v tv => tv = VUNIT
    | INT => fun _ v tv => 
      match v with | VN n => (tv = VINT n) | _ => False end
    | T1-->T2 => fun i f tf =>
      forall k v tv, k <= i -> irelLeVFun u srel T1 k v tv -> 
        MonadLe nil (irelLeVFun u srel T2) k (App::nil, v::f::nil) (APP tf tv)
    | T1***T2 => fun i v tv =>
      match v with 
        | VP v1 v2 => exists tv1, exists tv2, 
          tv = PAIR tv1 tv2 /\ 
          irelLeVFun u srel T1 i v1 tv1 /\ irelLeVFun u srel T2 i v2 tv2
        | _ => False
      end
    | T1+++T2 => fun i v tv =>
      match v with 
        | VP (VN (S _)) v' => exists tv', 
          tv = INL _ tv' /\ irelLeVFun u srel T1 i v' tv'
        | VP (VN 0) v' => exists tv', 
          tv = INR _ tv' /\ irelLeVFun u srel T2 i v' tv'
        | _ => False
      end
    | ALL T => fun i f tf =>
      forall t (tr: irel MValue (CVal t)), ireldecr tr ->
        MonadLe nil (irelLeVFun (S u) (isr_cons t tr srel) T)
        i (#v f) (<< SExp,PExp # irelLeV_C1 >> TAPP tf t)
    | EXIST T => fun i v tv =>
      exists ty, exists tr: irel MValue (CVal ty), exists tv', 
        ireldecr tr /\ tv = PACK tv' /\ 
        irelLeVFun (S u) (isr_cons ty tr srel) _ i v 
        (<< SVal,PVal # irelLeV_C1 >> tv')
  end.

Fixpoint irelLeVFun u srel T {struct T} := @irelLeVFun_body irelLeVFun u srel T.
Implicit Arguments irelLeVFun [u [T]].

Lemma irelLeVFun_fold: forall u srel T,
  @irelLeVFun_body irelLeVFun u srel T = @irelLeVFun u srel T.
Proof. by intros; destruct T. Qed.

Fixpoint irelLeEFun u (srel:iSubRel u) (E:Env u)
  : irel MEnv (EVal (subEnv (isr_sub srel) E)) 
  :=
  match E as env return irel MEnv (EVal (subEnv (isr_sub srel) env)) with 
    | nil => fun _ e _ => match e with | nil => True | _ => False end
    | T::TS => fun i e te =>
      match e with 
        | nil => False
        | v :: vs => 
          irelLeVFun srel i v (hdMap te) /\ irelLeEFun srel _ i vs (tlMap te)
      end
  end.
Implicit Arguments irelLeEFun [u [E]].

Definition irelLeCFun u srel E T 
  : irel MComp (Exp 0 (subEnv (isr_sub srel) E) (subTy (isr_sub srel) T)) 
  := fun i c tc =>
    forall k e te,
      k <= i -> @irelLeEFun u srel E k e te ->
      MonadLe e (@irelLeVFun u srel T) k c (SubExp te tc).
Implicit Arguments irelLeCFun [u [E] [T]].

(* irelLe *)

Definition irelLeV u T : irel MValue (Val u nil T) := fun i v tv =>
  forall srel, isr_decr srel -> 
    irelLeVFun srel i v (subVal (isr_sub srel) tv).

Definition irelLeE u (E:Env u) : irel MEnv (EVal E) := fun i e te =>
  forall srel, isr_decr srel ->
    irelLeEFun srel i e (subSub (isr_sub srel) te).

Definition irelLeC u E T : irel MComp (Exp u E T) := fun i c tc =>
  forall srel, isr_decr srel -> 
    irelLeCFun srel i c (subExp (isr_sub srel) tc).

(*
   relLe
*)

Definition relLeV u T := fun v tv =>
  exists tv', tv' <<V<< (cstr tv) /\ forall k, @irelLeV u T k v tv'.

Definition relLeE u E := fun e te =>
  exists te', te' <<E<< (cstr te) /\ forall k, @irelLeE u E k e te'.

Definition relLeC u E T := fun comp tc =>
  exists tc', tc' <<C<< (cstr tc) /\ forall k, @irelLeC u E T k comp tc'.

(*
   subiSR
*)

Definition reniSR u w (ren: renT u w) (srel: iSubRel w) : iSubRel u := 
  fun var => srel (ren var).

Definition subiSR u w (sub: subT u w) (srel: iSubRel w) : iSubRel u := 
  fun var => {{_# (isr_sub srel @ss@ sub) var , @irelLeVFun w srel (sub var) }}.

(**
   >> relation
 **)

(* SubRel *)

Definition SRel (ty:Ty 0) : Type := rel MValue (CVal ty).

Definition SubRel u := varT u -> sigT SRel.

Definition sr_sub u (srel:SubRel u) : subT u 0 := fun var => projT1 (srel var).

Definition sr_rel u (srel:SubRel u) := fun var => projT2 (srel var).

Definition sr_cons u t tr (srel: SubRel u) : SubRel (S u) :=
  @consmap (fun _ => sigT SRel) u 0 ({{ SRel # t , tr}}) srel.
Implicit Arguments sr_cons [u].

Definition sr0 : SubRel 0 := fun var => varT_empty _ var.

Lemma sr_sub_cons: forall u t tr (srel: SubRel u),
  sr_sub (sr_cons t tr srel) = consmap t (sr_sub srel).
Proof. intros; mapext_destr. Qed.

Lemma sr0_idsub: sr_sub sr0 = idsub.
Proof. mapext_destr. Qed.

Lemma sr_singsub: forall u (srel:SubRel u) ty,
  sr_sub srel @ss@ [|ty|] = consmap (subTy (sr_sub srel) ty) (sr_sub srel).
Proof. intros; mapext_destr. Qed.

(**
   biorthogonaltiy 
**)

Definition pton_termv (pe:MEnv) (P: prdc MValue) : prdc CESD := fun cesd =>
  match cesd with (c,e,s,d) =>
    forall val, P val -> 
      terminate (c, pe++e, val::s, d)
  end.

Definition ntop_termc (pe:MEnv) (P: prdc CESD) : prdc MComp := fun comp =>
  match comp with (c1,s1) =>
    forall c e s d, P (c,e,s,d) -> 
      terminate (c1++c, pe++e, s1++s, d)
  end.

Definition rel_to_pred (A B:Type) (r:rel A B) (b:B) : prdc A := fun a => r a b.

Definition valset_ge := rel_to_pred.

Definition MonadGe e T (RT: rel MValue (CVal T)) : rel MComp (CExp T) := 
  fun comp tc => 
    forall tv, tc |==> tv ->
      ntop_termc e (pton_termv e (valset_ge RT tv)) comp.

(* 
   Approximation relation
*)

(* erelGe Functor *)

Lemma erelGeV_C1 : forall u (srel: SubRel u) T t (tr: rel MValue (CVal t)),
  (nil:Env 0, (subTy [|t|] (subTy (subL (sr_sub srel)) T))) =
  (nil:Env 0, (subTy (sr_sub (sr_cons t tr srel)) T)).
Proof. by intros; simpl_sub; rewrite ! sr_sub_cons. Qed.
Implicit Arguments erelGeV_C1 [u srel T t tr].

Definition erelGeVFun_body
  (erelGeVFun: forall u (srel:SubRel u) (T:Ty u), 
    rel MValue (CVal (subTy (sr_sub srel) T)))
  u (srel:SubRel u) (T:Ty u) : rel MValue (CVal (subTy (sr_sub srel) T)) :=

  match T return rel MValue (CVal (subTy (sr_sub srel) T)) with
    | TYVAR var => projT2 (srel var)
    | UNIT => fun v tv => tv = VUNIT
    | INT => fun v tv => 
      match v with | VN n => (tv = VINT n) | _ => False end
    | T1-->T2 => fun f tf =>
      forall v tv, erelGeVFun u srel T1 v tv ->
        MonadGe nil (erelGeVFun u srel T2) (App::nil, v::f::nil) (APP tf tv)
    | T1***T2 => fun v tv =>
      match v with 
        | VP v1 v2 => exists tv1, exists tv2,
          tv = PAIR tv1 tv2 /\ 
          erelGeVFun u srel T1 v1 tv1 /\ erelGeVFun u srel T2 v2 tv2
        | _ => False
      end
    | T1+++T2 => fun v tv =>
      match v with 
        | VP (VN (S _)) v' => exists tv', 
          tv = INL _ tv' /\ erelGeVFun u srel T1 v' tv'
        | VP (VN 0) v' => exists tv', 
          tv = INR _ tv' /\ erelGeVFun u srel T2 v' tv'
        | _ => False
      end
    | ALL T => fun f tf =>
      forall t (tr: rel MValue (CVal t)),
        MonadGe nil (erelGeVFun (S u) (sr_cons t tr srel) T)
        (#v f) (<< SExp,PExp # erelGeV_C1 >>TAPP tf t)
    | EXIST T => fun v tv =>
      exists ty, exists tr: rel MValue (CVal ty), exists tv',
        tv = PACK tv' /\ 
        erelGeVFun (S u) (sr_cons ty tr srel) _ v
        (<< SVal,PVal # erelGeV_C1 >> tv')
  end.

Fixpoint erelGeVFun u srel T {struct T} := @erelGeVFun_body erelGeVFun u srel T.
Implicit Arguments erelGeVFun [u [T]].

Lemma erelGeVFun_fold: forall u srel T,
  @erelGeVFun_body erelGeVFun u srel T = @erelGeVFun u srel T.
Proof. by intros; destruct T. Qed.

Fixpoint erelGeEFun u (srel: SubRel u) (E:Env u)
  : rel MEnv (EVal (subEnv (sr_sub srel) E)) :=
  match E as env return rel MEnv (EVal (subEnv (sr_sub srel) env)) with
    | nil => (fun e _ => match e with | nil => True | _ => False end)
    | T::TS => fun e te =>
      match e with 
        | nil => False
        | v :: vs => 
          erelGeVFun srel v (hdMap te) /\ erelGeEFun srel _ vs (tlMap te)
      end
  end.
Implicit Arguments erelGeEFun [u E].

Definition erelGeCFun u (srel: SubRel u) (E:Env u) T
  : rel MComp (Exp _ (subEnv (sr_sub srel) E) (subTy (sr_sub srel) T)) 
  := fun c tc => forall e te,
    @erelGeEFun u srel E e te ->
    MonadGe e (@erelGeVFun u srel T) c (SubExp te tc).
Implicit Arguments erelGeCFun [u E T].

(* erelGe *)

Definition erelGeV u (T:Ty u) : rel MValue (Val _ nil T) := fun v tv =>
  forall srel, 
    erelGeVFun srel v (subVal (sr_sub srel) tv).

Definition erelGeE u (E:Env u) : rel MEnv (EVal E) := fun e te =>
  forall srel,
    erelGeEFun srel e (subSub (sr_sub srel) te).

Definition erelGeC u E T : rel MComp (Exp u E T) := fun c tc => 
  forall srel,
    erelGeCFun srel c (subExp (sr_sub srel) tc).

(* 
   relGe
*)

Definition relGeV u T := fun v tv =>
  exists tvs, tv <<V<< tvs /\ (forall p, @erelGeV u T v (tvs p)).

Definition relGeE u E := fun e te =>
  exists tes, te <<E<< tes /\ (forall p, @erelGeE u E e (tes p)).

Definition relGeC u E T : rel MComp (Exp u E T) := fun comp tc =>
  exists tcs, tc <<C<< tcs /\ (forall p, @erelGeC u E T comp (tcs p)).

(*
   == relations
*)

Definition relEqV u T v tv := @relLeV u T v tv /\ relGeV v tv.

Definition relEqE u E e te := @relLeE u E e te /\ relGeE e te.

Definition relEqC u E T c tc := @relLeC u E T c tc /\ relGeC c tc.

(* subSR *)

Definition renSR u w (ren: renT u w) (srel: SubRel w) : SubRel u := 
  fun var => srel (ren var).

Definition subSR u w (sub: subT u w) (srel: SubRel w) : SubRel u := 
  fun var => {{ _ # (sr_sub srel @ss@ sub) var , @erelGeVFun w srel (sub var) }}.

(*
   Notations
*)

Notation "v '~~' i '<vv<' tv" := (irelLeV i v tv) 
  (at level 70, no associativity).
Notation "e '~~' i '<ee<' te" := (irelLeE i e te) 
  (at level 70, no associativity).
Notation "c '~~' i '<cc<' tc" := (irelLeC i c tc)
  (at level 70, no associativity).

Notation "v '>vv>' tv" := (erelGeV v tv) (at level 70, no associativity).
Notation "e '>ee>' te" := (erelGeE e te) (at level 70, no associativity).
Notation "c '>cc>' tc" := (erelGeC c tc) (at level 70, no associativity).

Notation "v '<<vv<<' tv" := (relLeV v tv) (at level 70, no associativity).
Notation "e '<<ee<<' te" := (relLeE e te) (at level 70, no associativity).
Notation "c '<<cc<<' tc" := (relLeC c tc) (at level 70, no associativity).

Notation "v '>>vv>>' tv" := (relGeV v tv) (at level 70, no associativity).
Notation "e '>>ee>>' te" := (relGeE e te) (at level 70, no associativity).
Notation "c '>>cc>>' tc" := (relGeC c tc) (at level 70, no associativity).

Notation "v '==vv==' tv" := (relEqV v tv) (at level 70, no associativity).
Notation "e '==ee==' te" := (relEqE e te) (at level 70, no associativity).
Notation "c '==cc==' tc" := (relEqC c tc) (at level 70, no associativity).

(*
   Lemmas for <<
*)

Lemma pton_kstepv_contra: forall pe p q,
  subset p q -> subset (pton_kstepv pe q) (pton_kstepv pe p).
Proof. 
do 2 red; destruct a as [i [[[c e] s] d]]; intros.
by eapply H0; first by Apply; Apply.
Qed.

Lemma ntop_kstepc_contra: forall pe p q,
  subset p q -> subset (ntop_kstepc pe q) (ntop_kstepc pe p).
Proof.
do 2 red; destruct a as [i [c s]]; intros.
by eapply H0; first by Apply; Apply.
Qed.

Lemma pton_kstepv_decr: forall pe P, idecr (@pton_kstepv pe P).
Proof.
do 2 red; destruct a as [[[c e] s] d]; intros.
eapply H0; by [Apply | omega | ].
Qed.

Lemma ntop_kstepc_decr: forall pe P, idecr (@ntop_kstepc pe P).
Proof.
do 2 red; destruct a as [c s]; intros.
eapply H0; by [Apply | omega | ].
Qed.

Lemma irel_to_ipred_variance: forall A B (r r':irel A B) b b', 
  (forall k a, r k a b -> r' k a b') -> 
  subset (irel_to_ipred r b) (irel_to_ipred r' b').
Proof. by red; intros; destruct a; apply H. Qed.

Lemma MonadLe_covar: forall e T (RT RT': irel MValue (CVal T)),
  subset3 RT RT' -> subset3 (@MonadLe e _ RT) (@MonadLe e _ RT').
Proof.
do 2 red; intros; revert H0; unfold MonadLe.
apply ntop_kstepc_contra, pton_kstepv_contra, irel_to_ipred_variance.
intros; destruct H0; exists x; split; by [Apply | intuition].
Qed.

Lemma ivalset_le_decr: forall T RT exp,
  ireldecr RT -> idecr (@ivalset_le T RT exp).
Proof.
repeat red; intros; destruct H1 as [tv [? ?]].
exists tv; split; by [ | eapply H; Apply ].
Qed.

Lemma MonadLe_decr: forall e T RT, ireldecr (@MonadLe e T RT).
Proof. do 2 red; intros; eapply ntop_kstepc_decr; Apply. Qed.

Lemma isr_cons_decr: forall u t tr (srel:iSubRel u),
  ireldecr tr -> isr_decr srel -> isr_decr (@isr_cons _ t tr srel).
Proof. by red; intros; depdestr var; Apply. Qed.

Lemma irelLeVFun_arr_zero: forall u srel T1 T2 tv tv',
  @irelLeVFun u srel (T1-->T2) 0 tv tv'.
Proof. do 4 red; intros; replace k0 with 0 by omega; apply stepsfor_zero. Qed.

Lemma irelLeVFun_decr: forall u T srel,
  isr_decr srel -> ireldecr (@irelLeVFun u srel T).
Proof.
induction T; red; simpl; intros; auto; fold_Sub. 
- by apply H with k'; [ omega | auto ].
- by apply H1; [ omega | auto ].
- destruct a; auto.
  destruct H1 as [tv1 [tv2 [? [? ?]]]]; exists tv1; exists tv2. 
  split; first by []; split.
  - eapply IHT1; first (by []); first (by Apply); by Apply.
  - eapply IHT2; first (by []); first (by Apply); by Apply.
- destruct a; auto; destruct a1; auto. 
  destruct n; destruct H1 as [tv' ?]; exists tv'; intuition.
  - by eapply IHT2; Apply.
  - by eapply IHT1; Apply.
- apply MonadLe_decr with k'; by [omega | Apply].
- destruct H1 as [t [tr [tv [? [? ?]]]]]; exists t; exists tr; exists tv.
  split; first by []; split; first by [].
  eapply IHT.
  - by red; intros; depdestr var; Apply.
  - apply H0.
  - eapply IHT; by [ apply isr_cons_decr | apply le_n | ].
Qed.

Lemma irelLeEFun_decr: forall u srel E,
  isr_decr srel -> ireldecr (@irelLeEFun u srel E).
Proof.
induction E; red; intros; simpl; first by [].
destruct a0; auto; split.
- by eapply irelLeVFun_decr; Apply.
- by eapply IHE; Apply.
Qed.

Lemma irelLeCFun_decr: forall u srel (E:Env u) T,
  ireldecr (@irelLeCFun u srel E T).
Proof. do 2 red; intros; apply H0; by [ omega | ]. Qed.

Lemma irelLeEFun_len: forall u srel E k e te, 
  @irelLeEFun u srel E k e te -> length e = length E.
Proof.
induction E; intros.
- destruct e; by [ | inversion H].
- destruct e; [inversion H | simpl; f_equal; eapply IHE; Apply].
Qed.
Implicit Arguments irelLeEFun_len [ u srel E k e te ].

Lemma irelLeEFun_inv: forall u srel E k e te T (var: @VarT u E T), 
  irelLeEFun srel k e te ->
  exists v, nth_error e (idxVar var) = value v /\ 
    @irelLeVFun u srel T k v (te _ (subVar (isr_sub srel) var)).
Proof.
induction E; intros; first by inversion var.
destruct e; first by inversion H. 
depdestr var.
- exists m; split; by [ | Apply].
- destruct (IHE k e (tlMap te) _ var) as [v [? ?]]; first by Apply.
  exists v; split; by [ | Apply].
Qed.

Lemma MonadLe_val: forall e T (RT:irel MValue (CVal T)) i v tc tv,
  RT i v tv -> tc |==> tv -> @MonadLe e T RT i (#v v) tc.
Proof.
do 2 red; unfold VtoC; intros; simpl. 
eapply H1; by [ exists tv; split; [ | Apply] | | ].
Qed.

Lemma relLeC_le: forall u E T c (tc tc': Exp u E T),
  tc <<C<< (cstr tc') -> c <<cc<< tc -> c <<cc<< tc'.
Proof.
intros; destruct H0; exists x.
split; by [eapply ObsLeC_trans; intros; Apply | Apply].
Qed.

Lemma subiSR_singsub_isrel: forall u ty (srel: iSubRel u),
  subiSR ([|ty|]) srel = isr_cons _ (@irelLeVFun u srel ty) srel.
Proof.
by intros; extension var; apply Heq_sigT; depdestr var.
Qed.

Lemma subiSR_shsub_cons: 
  forall u ty (tr: irel MValue (CVal ty)) (srel:iSubRel u),
  subiSR shsub (isr_cons ty tr srel) = srel.
Proof. by intros; extension var; apply Heq_sigT. Qed.

Lemma reniSR_renL: 
  forall u w (ren:renT u w) ty (tr: irel MValue (CVal ty)) (srel: iSubRel w), 
  reniSR (renL ren) (isr_cons ty tr srel) = isr_cons ty tr (reniSR ren srel).
Proof.
by intros; extension var; apply Heq_sigT; depdestr var.
Qed.

Lemma irelLeVFun_ren: forall u T w (ren:renT u w) (srel: iSubRel w),
  {{SirelV # 
    @irelLeVFun _ srel (renTy ren T) == @irelLeVFun _ (reniSR ren srel) T}}.
Proof.
induction T; intros; auto; simpl; fold_Sub; 
 try (Hrewritec <- (IHT1 _ ren srel); Hrewritec <- (IHT2 _ ren srel); auto).
- do 2 (apply (Heq_extensionality SirelV SirelV 
               (fun ty P i f tf => forall t, P t i f tf)); 
        [by left; simpl_sub | intros]).
  Hsimpl; Hrewritec (IHT _ (renL ren) (isr_cons a a0 srel)).
  by simpl_sub; rewrite -> reniSR_renL, (subTy_ss (subL _)); simpl_sub; Hunify.
- do 2 (apply (Heq_extensionality SirelV SirelV 
               (fun ty P i f tf => exists t, P t i f tf)); 
        [by left; simpl_sub | intros]).
  Hsimpl; Hrewritec (IHT _ (renL ren) (isr_cons a a0 srel)).
  by simpl_sub; rewrite -> reniSR_renL, (subTy_ss (subL _)); simpl_sub; Hunify.
Qed.

Lemma subiSR_subL: 
  forall u w (sub:subT u w) ty (tr: irel MValue (CVal ty)) (srel: iSubRel w), 
  subiSR (subL sub) (isr_cons ty tr srel) = isr_cons ty tr (subiSR sub srel).
Proof.
intros; extension var; apply Heq_sigT; depdestr var;
 first by [].
simpl; Hrewritec (irelLeVFun_ren (sub var) shren (isr_cons ty tr srel)).
replace (reniSR shren (isr_cons ty tr srel)) with srel; first by [].
by extension var'; depdestr var'.
Qed.

Lemma irelLeVFun_sub: forall u T w (sub:subT u w) (srel: iSubRel w),
  {{SirelV # 
    @irelLeVFun _ srel (subTy sub T) == @irelLeVFun _ (subiSR sub srel) T}}.
Proof.
induction T; intros; auto; simpl; fold_Sub;
  try (Hrewritec <- (IHT1 _ sub srel); Hrewritec <- (IHT2 _ sub srel); auto).
- do 2 (apply (Heq_extensionality SirelV SirelV 
               (fun ty P i f tf => forall t, P t i f tf)); 
        [by left; simpl_sub | intros]).
  Hsimpl; Hrewritec (IHT _ (subL sub) (isr_cons a a0 srel)).
  by simpl_sub; rewrite -> subiSR_subL, (subTy_ss (subL _)); simpl_sub; Hunify.
- do 2 (apply (Heq_extensionality SirelV SirelV 
               (fun ty P i f tf => exists t, P t i f tf)); 
        [by left; simpl_sub | intros]).
  Hsimpl; Hrewritec (IHT _ (subL sub) (isr_cons a a0 srel)).
  by simpl_sub; rewrite -> subiSR_subL, (subTy_ss (subL _)); simpl_sub; Hunify.
Qed.

Lemma irelLeEFun_sub: forall u E w (sub:subT u w) (srel: iSubRel w),
  {{SirelE #
    @irelLeEFun w srel (subEnv sub E) == @irelLeEFun u (subiSR sub srel) E}}.
Proof.
induction E; intros; first by [].
simpl; fold_Sub; Hrewritec (irelLeVFun_sub a sub srel).
by Hrewritec (IHE _ sub srel).
Qed.

(*
   Lemmas for >>
*)

Lemma pton_termv_contra: 
  forall pe p q, subset p q -> subset (pton_termv pe q) (pton_termv pe p).
Proof.
do 2 red; destruct a as [[[c e] s] d]; intros; by eapply H0; Apply; Apply.
Qed.

Lemma ntop_termc_contra: forall pe p q,
  subset p q -> subset (ntop_termc pe q) (ntop_termc pe p).
Proof. do 2 red; destruct a as [c s]; intros; by eapply H0; Apply; Apply. Qed.

Lemma rel_to_pred_variance: forall A B (r r':rel A B) b b', 
  (forall a, r a b -> r' a b') -> subset (rel_to_pred r b) (rel_to_pred r' b').
Proof. by red; intros; apply H. Qed.

Lemma MonadGe_covar: forall e T (RT RT': rel MValue (CVal T)),
  subset2 RT RT' -> subset2 (@MonadGe e _ RT) (@MonadGe e _ RT').
Proof.
do 2 red; intros; move: (H0 tv H1).
apply ntop_termc_contra, pton_termv_contra, rel_to_pred_variance. 
by intros; Apply.
Qed.

Lemma erelGeEFun_len: forall u srel E e te, 
  @erelGeEFun u srel E e te -> length e = length E.
Proof.
induction E; intros.
  destruct e; by [ | inversion H].
  destruct e; by [ inversion H | simpl; f_equal; eapply IHE; apply H].
Qed.
Implicit Arguments erelGeEFun_len [ u srel E e te ].

Lemma erelGeEFun_inv: forall u srel E e te T (var:@VarT u E T), 
  @erelGeEFun u srel E e te ->
  exists v, nth_error e (idxVar var) = value v /\ 
    @erelGeVFun u srel T v (te _ (subVar (sr_sub srel) var)).
Proof.
induction E; intros; first by inversion var.
destruct e; first by inversion H.  
depdestr var.
- exists m; split; by [ | Apply]. 
- destruct (IHE e (tlMap te) _ var) as [v [? ?]]; first by apply H.
  exists v; split; by [ | Apply].
Qed.

Lemma MonadGe_val: forall e T (RT:rel MValue (CVal T)) v tc tv,
  RT v tv -> tc |==> tv -> @MonadGe e T RT (#v v) tc.
Proof.
do 2 red; unfold VtoC; intros; simpl; eapply H2.
do 2 red; destruct H0; destruct H1; destruct (Ev_uniq H0 H1). 
by rewrite <- H4.
Qed.

Lemma BotF_bot: forall v u srel A B,
  @erelGeVFun u srel (A -->B) v 
  (BotF nil (subTy (sr_sub srel) A) (subTy (sr_sub srel) B)).
Proof.
do 4 red; intros; fold erelGeVFun in *.
destruct H0 as [i ?]; replace_hyp H0 (Ev_to_Treduce H0).
destruct (BotF_diverge (S i) (subTy (sr_sub srel) B) tv) as [i' [tv' [? ?]]].
replace i' with (i+(i'-i)) in H3 by omega.
destruct (Treduce_VAL_inv (Treduce_inv H0 _ H3)).
replace i' with i in H2 by omega; contradiction (@le_Sn_n _ H2).
Qed.

Lemma subSR_singsub_srel: forall u ty (srel: SubRel u),
  subSR [|ty|] srel = sr_cons _ (@erelGeVFun _ srel ty) srel.
Proof.
by intros; extension var; apply Heq_sigT; depdestr var.
Qed.

Lemma subSR_shsub_cons: forall u ty (tr: rel MValue (CVal ty)) (srel:SubRel u),
  subSR shsub (sr_cons ty tr srel) = srel.
Proof. by intros; extension var; apply Heq_sigT. Qed.

Lemma renSR_renL: 
  forall u w (ren:renT u w) ty (tr: rel MValue (CVal ty)) (srel: SubRel w), 
  renSR (renL ren) (sr_cons ty tr srel) = sr_cons ty tr (renSR ren srel).
Proof.
by intros; extension var; apply Heq_sigT; depdestr var.
Qed.

Lemma erelGeVFun_ren: forall u T w (ren:renT u w) (srel: SubRel w),
  {{SrelV # 
    @erelGeVFun _ srel (renTy ren T) == @erelGeVFun _ (renSR ren srel) T}}.
Proof.
induction T; intros; auto; simpl; fold_Sub;
  try (Hrewritec <- (IHT1 _ ren srel); Hrewritec <- (IHT2 _ ren srel); auto).
- do 2 (apply (Heq_extensionality SrelV SrelV 
               (fun ty P f tf => forall t, P t f tf)); 
        [by left; simpl_sub | intros]).
  Hsimpl; Hrewritec (IHT _ (renL ren) (sr_cons a a0 srel)).
  by simpl_sub; rewrite -> renSR_renL, (subTy_ss (subL _)); simpl_sub; Hunify.
- do 2 (apply (Heq_extensionality SrelV SrelV 
               (fun ty P f tf => exists t, P t f tf)); 
        [by left; simpl_sub | intros]).
  Hsimpl; Hrewritec (IHT _ (renL ren) (sr_cons a a0 srel)).
  by simpl_sub; rewrite -> renSR_renL, (subTy_ss (subL _)); simpl_sub; Hunify.
Qed.

Lemma subSR_subL:
  forall u w (sub:subT u w) ty (tr: rel MValue (CVal ty)) (srel: SubRel w), 
  subSR (subL sub) (sr_cons ty tr srel) = sr_cons ty tr (subSR sub srel).
Proof.
intros; extension var; apply Heq_sigT; depdestr var; 
 first by [].
simpl; Hrewritec (erelGeVFun_ren (sub var) shren (sr_cons ty tr srel)).
replace (renSR shren (sr_cons ty tr srel)) with srel; first by [].
by extension var'; depdestr var'.
Qed.

Lemma erelGeVFun_sub: forall u T w (sub:subT u w) (srel: SubRel w),
  {{SrelV # 
    @erelGeVFun _ srel (subTy sub T) == @erelGeVFun _ (subSR sub srel) T}}.
Proof.
induction T; intros; auto; simpl; fold_Sub;
  try (Hrewritec <- (IHT1 _ sub srel); Hrewritec <- (IHT2 _ sub srel); auto).
- do 2 (apply (Heq_extensionality SrelV SrelV 
               (fun ty P f tf => forall t, P t f tf)); 
        [by left; simpl_sub | intros]).
  Hsimpl; Hrewritec (IHT _ (subL sub) (sr_cons a a0 srel)).
  by simpl_sub; rewrite -> subSR_subL, (subTy_ss (subL _)); simpl_sub; Hunify.
- do 2 (apply (Heq_extensionality SrelV SrelV 
               (fun ty P f tf => exists t, P t f tf)); 
        [by left; simpl_sub | intros]).
  Hsimpl; Hrewritec (IHT _ (subL sub) (sr_cons a a0 srel)).
  by simpl_sub; rewrite -> subSR_subL, (subTy_ss (subL _)); simpl_sub; Hunify.
Qed.

Lemma erelGeEFun_sub: forall u E w (sub:subT u w) (srel: SubRel w),
  {{SrelE #
    @erelGeEFun w srel (subEnv sub E) == @erelGeEFun u (subSR sub srel) E}}.
Proof.
induction E; intros; first by []. 
simpl; fold_Sub; Hrewritec (erelGeVFun_sub a sub srel). 
by Hrewritec (IHE _ sub srel).
Qed.

(*
   Compositionality of Logical Relation
*)

Definition CApp (cf cx:Code) := cf++cx++[App].

Lemma compositionality_le: 
  forall u E A B f x (tf:Val u E (A-->B)) (tx:Val u E A),
  #c f <<cc<< ^tf -> #c x <<cc<< ^tx -> #c (CApp f x) <<cc<< APP tf tx.
Proof.
move => u E A B f x tf tx [tf' [Htfsrc Htf]] [tx' [Htxsrc Htx]].
exists (LET tf' (LET (SubExp (shSub _) tx') 
 (APP (VAR (SVAR _ ZVAR)) (VAR ZVAR)))).
split.
- apply ObsLeC_trans with (cstr (LET (^tf) (LET (SubExp (shSub _) (^tx)) 
   (APP (VAR (SVAR _ ZVAR)) (VAR ZVAR))))).
  - apply ObsLe_LET; first by [].
    apply ObsLe_LET.
    - by apply ObsLeC_Sub; [intros; apply ObsLeV_refl | ].
    - by apply ObsLe_APP; apply ObsLeV_refl.
  - move => p; eapply ObsLeC_trans with (cstr _).
    - eapply ObsLe_beta.
    - move => p'; unfold_SubVE; eapply ObsLeC_trans with (cstr _).
      - eapply ObsLe_beta.
      - move => p''; unfold_SubVE; simpl_Sub; apply ObsLeC_refl.
- move => k srel Hsrel k0 e te Hk0k Hete j c0 e0 s0 d0 Hcont0 k1 Hk1k0 Hk1j.
  unfold CApp; rewrite !app_ass.
  eapply Htf; [by Apply |by apply Hk0k |by Apply | | | ].
  instantiate (1:=k1).
  move => i v [tf'' [[n1 Htf''] Hv]] k2 Hk2i Hk2k1.
  eapply Htx; [by Apply |by apply Hk2k1 | | | | ].
  - by eapply irelLeEFun_decr; last (by Apply); first (by Apply); by omega.
  - instantiate (1:=k2).
    move => i0 v0 [tx'' [[n2 Htx''] Hv0]] k3 Hk3i0 Hk3k2.
    change (v0::v::s0) with ([v0;v]++s0).
    eapply Hv; fold irelLeVFun (app e).
    - instantiate (1:=k3); omega.
    - by eapply irelLeVFun_decr; last (by Apply); [Apply | omega].
    - instantiate (1:=k3).
      move => i1 v1 [tftx [[n3 Htftx] Hv1]] k4 Hk4i1 Hk4k3.
      eapply Hcont0.
      eexists; split.
      - eexists.
        unfold_SubVE; econstructor; first by Apply. 
        unfold_SubVE; econstructor; first by do 2 simpl_Sub; Apply.
        by unfold_SubVE; simpl_Sub; Apply.
      - by Apply.
omega. omega . omega. omega. omega. omega. omega. omega.
Qed.

Lemma compositionality_ge:
  forall u E A B f x (tf:Val u E (A-->B)) (tx:Val u E A),
  #c f >>cc>> ^tf -> #c x >>cc>> ^tx -> #c (CApp f x) >>cc>> APP tf tx.
Proof.
move => u E A B f x tf tx [tfs [Htfsrc Hftf]] [txs [Htxsrc Hxtx]].
exists (fun i => LET (tfs i) (LET (SubExp (shSub _) (txs i)) 
 (APP (VAR (SVAR _ ZVAR)) (VAR ZVAR)))).
split.
- apply ObsLeC_trans with (cstr (LET (^tf) (LET (SubExp (shSub _) (^tx)) 
   (APP (VAR (SVAR _ ZVAR)) (VAR ZVAR))))); first last.
  - move => p; apply ObsLe_LET; first by [].
    apply ObsLe_LET; first by apply ObsLeC_Sub; [intros; apply ObsLeV_refl | ].
    by apply ObsLe_APP; apply ObsLeV_refl.
  - eapply ObsLeC_trans with (cstr _); first last.
    - move => p; eapply ObsLe_beta.
    - unfold_SubVE; eapply ObsLeC_trans with (cstr _); first last. 
      - move => p; eapply ObsLe_beta.
      - unfold_SubVE; simpl_Sub; apply ObsLeC_refl.
- move => p srel e te Hete tv.
  unfold_SubVE; move => [k Htv].
  inversion Htv; Hsubst; clear Htv.
  move : H3 H6; unfold_SubVE; do 2 simpl_Sub; move : v1 => tv0 Htv0 Htv.
  inversion Htv; Hsubst; clear Htv.
  move : H3 H6; unfold_SubVE; simpl_Sub; move : v1 => tv1 Htv1 Htv.
  move => c0 e0 s0 d0 Hcont0.
  unfold CApp; rewrite !app_ass.
  eapply (Hftf p); [by Apply | by eexists; Apply | ].
  move => v Hv.
  change (v::s0) with ([v]++s0).
  eapply (Hxtx p); [by Apply | by eexists; Apply | ].
  move => v0 Hv0.
  change (v0::v::s0) with ([v0;v]++s0).
  eapply Hv; [by Apply | by eexists; Apply | ].
  by simpl; Apply.
Qed.

Corollary compositionality:
  forall u E A B f x (tf:Val u E (A-->B)) (tx:Val u E A),
  #c f ==cc== ^tf -> #c x ==cc== ^tx -> #c (CApp f x) ==cc== APP tf tx.
Proof.
by intros; split; 
 [ apply compositionality_le | apply compositionality_ge ]; Apply.
Qed.

(****
   Full abstraction
 ****)

Theorem relLeV_preserve_ObsLeV: forall u T v (tv tv': Val u nil T),
  v <<vv<< tv -> tv <<V<< cstr tv' -> v <<vv<< tv'.
Proof.
by intros; destruct H as [tv'' [? ?]]; exists tv''; split;
 [ eapply ObsLeV_trans; intros; Apply | ].
Qed.

Theorem relGeV_preserve_ObsLeV: forall u T v (tv tv': Val u nil T),
  v >>vv>> tv -> tv' <<V<< cstr tv -> v >>vv>> tv'.
Proof.
by intros; destruct H as [tvs [? ?]]; exists tvs; split;
 [ eapply ObsLeV_trans; intros; Apply | ].
Qed.  

Theorem relLeE_preserve_ObsLeE: forall u (E:Env u) e (te te': EVal E),
  e <<ee<< te -> te <<E<< cstr te' -> e <<ee<< te'.
Proof.
by intros; destruct H as [te'' [? ?]]; exists te''; split;
 [ eapply ObsLeE_trans; intros; Apply | ].
Qed.

Theorem relGeE_preserve_ObsLeE: forall u (E:Env u) e (te te': EVal E),
  e >>ee>> te -> te' <<E<< cstr te -> e >>ee>> te'.
Proof.
by intros; red in H; destruct H as [tes [? ?]]; exists tes; split;
 [ eapply ObsLeE_trans; intros; Apply | ].
Qed.

Theorem relLeC_preserve_ObsLeC: forall u E T c (tc tc': Exp u E T),
  c <<cc<< tc -> tc <<C<< cstr tc' -> c <<cc<< tc'.
Proof.
by intros; destruct H as [tc'' [? ?]]; exists tc''; split;
 [ eapply ObsLeC_trans; intros; Apply | ].
Qed.

Theorem relGeC_preserve_ObsLeC: forall u E T c (tc tc': Exp u E T),
  c >>cc>> tc -> tc' <<C<< cstr tc -> c >>cc>> tc'.
Proof.
by intros; destruct H as [tcs [? ?]]; exists tcs; split;
 [ eapply ObsLeC_trans; intros; Apply | ].
Qed.  

(****
   Adequacy of Logical Relation
****)

Definition comp_converge_to_int comp e n := 
  forall cesd,
  (terminate (#v(VN n) +~+ e +-+ cesd) ->
    terminate (comp +~+ e +-+ cesd))
  /\
  (diverge (#v(VN n) +~+ e +-+ cesd) ->
    diverge (comp +~+ e +-+ cesd)).

Definition comp_converge_to_unit comp e :=
  forall cesd,
  ((forall v, terminate (#v v +~+ e +-+ cesd)) ->
    terminate (comp +~+ e +-+ cesd))
  /\
  ((forall v, diverge (#v v +~+ e +-+ cesd)) -> 
    diverge (comp +~+ e +-+ cesd)).

Definition comp_diverge comp e :=
  forall cesd, diverge (comp +~+ e +-+ cesd).

Notation "c @ e ~>N n" := (comp_converge_to_int c e n)
  (at level 70, no associativity).
Notation "c @ e ~>" := (comp_converge_to_unit c e)
  (at level 70, no associativity).
Notation "c @ e ~>!"   := (comp_diverge c e)
  (at level 70, no associativity).

Lemma relLeC_adeq_int_bot:
  forall E (e:MEnv) (c:MComp) (te: EVal E) (tc: Exp _ E INT),
  e <<ee<< te -> c <<cc<< tc -> SubExp te tc |==>! -> c @ e ~>! .
Proof.
do 2 red; intros; destruct c as [c s]; destruct cesd as [[[c0 e0] s0] d0];
 destruct H0 as [tc' [? ?]]; destruct H as [te' [? ?]].
simpl; eapply H2.
- by instantiate (1:= isr0); red; intros; inversion var.
- by apply (le_n k).
- by apply H3; red; intros; inversion var.
- rewrite isr0_idsub; instantiate (1:= k).
  red; intros; destruct H4 as [tvv [? ?]].
  revert H4; Hrewrite (subSub_idsub te'); Hrewritec (subExp_idsub tc'); intro.
  assert (H8:= H4); destruct H4 as [j ?].
  assert (SubExp te' tc' <<C<< cstr (SubExp te tc))
   by apply (ObsLeC_Sub _ _ (cstr te) _ _ H H0).
  destruct (ObsLe_adeq_int _ _ H9 H8) as [? [j' ?]].
  destruct (H1 (S j')) as [j'' [tvv'' [? ?]]].
  replace j'' with (j'+(j''-j')) in H12 by omega.
  destruct (Treduce_VAL_inv (Treduce_inv (Ev_to_Treduce H10) _ H12)).
  by replace j'' with j' in H11 by omega; contradiction (le_Sn_n _ H11).
- by omega. - by omega.
Qed.

Lemma relLeC_adeq_unit_bot:
  forall E (e:MEnv) (c:MComp) (te: EVal E) (tc: Exp _ E UNIT),
  e <<ee<< te -> c <<cc<< tc -> SubExp te tc |==>! -> c @ e ~>! .
Proof.
do 2 red; intros; destruct c as [c s]; destruct cesd as [[[c0 e0] s0] d0];
 destruct H0 as [tc' [? ?]]; destruct H as [te' [? ?]].
simpl; eapply H2.
- by instantiate (1:= isr0); red; intros; inversion var.
- by apply (le_n k).
- by apply H3; red; intros; inversion var.
- rewrite isr0_idsub; instantiate (1:= k). 
  red; intros; destruct H4 as [tvv [? ?]].
  revert H4; Hrewrite (subSub_idsub te'); Hrewritec (subExp_idsub tc'); intro.
  assert (H8:= H4); destruct H4 as [j ?].
  assert (SubExp te' tc' <<C<< cstr (SubExp te tc))
   by (apply (ObsLeC_Sub _ _ (cstr te) _ _ H H0)).
  destruct (ObsLe_adeq_unit _ _ H9 H8) as [? [j' ?]]. 
  destruct (H1 (S j')) as [j'' [tvv'' [? ?]]].
  replace j'' with (j'+(j''-j')) in H12 by omega.
  destruct (Treduce_VAL_inv (Treduce_inv (Ev_to_Treduce H10) _ H12)).
  by replace j'' with j' in H11 by omega; contradiction (le_Sn_n _ H11).
- by omega. - by omega.
Qed.

Lemma relLeC_adeq_int_val:
  forall E (e:MEnv) (c:MComp) (te: EVal E) (tc: Exp _ E INT) n,
  e <<ee<< te -> c <<cc<< tc -> SubExp te tc |==> VINT n ->
  forall cesd, diverge (#v (VN n) +~+ e +-+ cesd) -> 
  diverge (c +~+ e +-+ cesd).
Proof.
red; intros; rename H2 into Hdiv.
destruct c as [c s]; destruct cesd as [[[c0 e0] s0] d0];
 destruct H0 as [tc' [? ?]]; destruct H as [te' [? ?]].
simpl; eapply H2.
- by instantiate (1:= isr0); red; intros; inversion var.
- by apply (le_n k).
- by apply H3; red; intros; inversion var.
- rewrite isr0_idsub; instantiate (1:= k). 
  red; intros; destruct H4 as [tvv [? ?]].
  revert H4; Hrewrite (subSub_idsub te'); Hrewritec (subExp_idsub tc'); intro.
  assert (H8:= H4); destruct H4 as [j ?].
  assert (SubExp te' tc' <<C<< cstr (SubExp te tc))
   by (apply (ObsLeC_Sub _ _ (cstr te) _ _ H H0)).
  destruct (ObsLe_adeq_int _ _ H9 H8) as [? [j' ?]]; destruct H1. 
  destruct (Ev_uniq H1 H10); destruct val; try contradiction H7.
  by simpl in H7; rewrite H7 in H12; depdestr H12; apply Hdiv.
- by omega. - by omega.
Qed.

Lemma relLeC_adeq_unit_val: 
  forall E (e:MEnv) (c:MComp) (te: EVal E) (tc: Exp _ E UNIT),
  e <<ee<< te -> c <<cc<< tc -> SubExp te tc |==> VUNIT ->
  forall cesd, 
    (forall v, diverge (#v v +~+ e +-+ cesd)) -> 
    diverge (c +~+ e +-+ cesd).
Proof.
red; intros; rename H2 into Hdiv.
destruct c as [c s]; destruct cesd as [[[c0 e0] s0] d0];
 destruct H0 as [tc' [? ?]]; destruct H as [te' [? ?]].
simpl; eapply H2.
- by instantiate (1:= isr0); red; intros; inversion var.
- by apply (le_n k).
- by apply H3; red; intros; inversion var.
- rewrite isr0_idsub; instantiate (1:= k). 
  red; intros; destruct H4 as [tvv [? ?]].
  revert H4; Hrewrite (subSub_idsub te'); Hrewritec (subExp_idsub tc'); intro.
  assert (H8:= H4); destruct H4 as [j ?].
  assert (SubExp te' tc' <<C<< cstr (SubExp te tc))
   by (apply (ObsLeC_Sub _ _ (cstr te) _ _ H H0)).
  destruct (ObsLe_adeq_unit _ _ H9 H8) as [? [j' ?]]. 
  destruct H1; destruct (Ev_uniq H1 H10).
  by apply (Hdiv val).
- by omega. - by omega.
Qed.

Lemma relGeC_adeq_int_val:
  forall E (e:MEnv) (c:MComp) (te: EVal E) (tc: Exp _ E INT) n,
  e >>ee>> te -> c >>cc>> tc -> SubExp te tc |==> VINT n ->
  forall cesd, 
    terminate (#v (VN n) +~+ e +-+ cesd) ->
    terminate (c +~+ e +-+ cesd).
Proof.
intros; rename H2 into Hterm.
destruct c as [c s]; destruct cesd as [[[c0 e0] s0] d0]. 
destruct H0 as [tcs [? ?]]; destruct H as [tes [? ?]].
simpl; assert (H6:= H1); destruct H6 as [j ?].
assert (SubExp te tc <<C<< fun i => SubExp (tes i) (tcs i))
 by (apply (ObsLeC_Sub _ _ tes _ _ H H0)).
destruct (ObsLe_adeq_int _ _ H5 H1) as [? [j' ?]]; eapply H2.
- by instantiate (2:= sr0); apply (H3 x).
- rewrite sr0_idsub; instantiate (2:= x). 
  Hhide_evars; Hrewrite (subSub_idsub (tes x));
   Hrewritec (subExp_idsub (tcs x)); Hshow_evars.
  by eexists; apply H6.
- red; intros; destruct val; try inversion H7.
  by repeat red in H7; depdestr H7; apply Hterm.
Qed.

Lemma relGeC_adeq_unit_val: 
  forall E (e:MEnv) (c:MComp) (te: EVal E) (tc: Exp _ E UNIT),
  e >>ee>> te -> c >>cc>> tc -> SubExp te tc |==> VUNIT ->
    forall cesd, 
      (forall v, terminate (#v v +~+ e +-+ cesd)) -> 
      terminate (c +~+ e +-+ cesd).
Proof.
intros; rename H2 into Hterm.
destruct c as [c s]; destruct cesd as [[[c0 e0] s0] d0]. 
destruct H0 as [tcs [? ?]]; destruct H as [tes [? ?]].
simpl; assert (H6:= H1); destruct H6 as [j ?].
assert (SubExp te tc <<C<< fun i => SubExp (tes i) (tcs i)) 
 by (apply (ObsLeC_Sub _ _ tes _ _ H H0)).
destruct (ObsLe_adeq_unit _ _ H5 H1) as [? [j' ?]]; eapply H2.
- by instantiate (2:= sr0); apply (H3 x).
- rewrite sr0_idsub; instantiate (2:= x). 
  Hhide_evars; Hrewrite (subSub_idsub (tes x));
   Hrewritec (subExp_idsub (tcs x)); Hshow_evars.
  by eexists; apply H6.
- by red; intros; apply (Hterm val).
Qed.

Corollary adequacy_int_bot:
  forall E (e:MEnv) (c:MComp) (te: EVal E) (tc: Exp _ E INT),
  e ==ee== te -> c ==cc== tc -> SubExp te tc |==>! -> c @ e ~>! .
Proof. by intros; eapply relLeC_adeq_int_bot; [apply H | apply H0 | ]. Qed.

Corollary adequacy_bool_bot: 
  forall E (e:MEnv) (c:MComp) (te: EVal E) (tc: Exp _ E UNIT),
  e ==ee== te -> c ==cc== tc -> SubExp te tc |==>! -> c @ e ~>! .
Proof. by intros; eapply relLeC_adeq_unit_bot; [apply H | apply H0 | ]. Qed.

Corollary adequacy_int_val:
  forall E (e:MEnv) (c:MComp) (te: EVal E) (tc: Exp _ E INT) n,
  e ==ee== te -> c ==cc== tc -> SubExp te tc |==> VINT n -> c @ e ~>N n.
Proof.
by intros; split; intros; 
 [ eapply relGeC_adeq_int_val | eapply relLeC_adeq_int_val ]; Apply.
Qed.

Corollary adequacy_bool_val:
  forall E (e:MEnv) (c:MComp) (te: EVal E) (tc: Exp _ E UNIT),
  e ==ee== te -> c ==cc== tc -> SubExp te tc |==> VUNIT -> c @ e ~>.
Proof.
by intros; split; intros;
 [ eapply relGeC_adeq_unit_val | eapply relLeC_adeq_unit_val ]; Apply.
Qed.

End LogicalRelationFunctor.

