(* 
   Standard operational semantics for \Fv
   Part of the formalization for
   Realizability and Compositional Compiler Correctness 
      for a Polymorphic Language
   N Benton & C-K Hur
   April 2010
*)
Require Import List Program EqNat Omega Bool_nat Heq.
Require Import util sysf sysfopsem.
Require Import ssreflect ssrbool Heq.

Set Implicit Arguments.

(* 
   SysF related definitions and lemmas 
*)

Definition SEnvTy u : Type := (Env u * Ty u) % type.
Definition SSVal (x: sigT SEnvTy) : Type := 
  let (u,envty) := x in let (env,ty) := envty in Val u env ty.
Definition SSExp (x: sigT SEnvTy) : Type := 
  let (u,envty) := x in let (env,ty) := envty in Exp u env ty.

Definition SEnvTyC: forall u w E T (C: u = w),
  {{SEnvTy # u, (E,T) }} =  {{SEnvTy # w, (<<Env#C>>E,<<Ty#C>>T)}}.
Proof. by intros; case C. Defined.
Implicit Arguments SEnvTyC [u w E T].

Fixpoint lifted_shren u w : renT (u+w) (S u + w) :=
  match u with
    | 0 => shren
    | S u' => renL (lifted_shren u' w)
  end.

Fixpoint lifted_shsub u w : subT (u+w) (S u + w) :=
  match u with
    | 0 => shsub
    | S u' => subL (lifted_shsub u' w)
  end.

Fixpoint lifted_shRen u E0 T (E:Env u) : RenT (E0++E) (E0++T::E) :=
  match E0 with
    | nil => shRen T
    | T0::E0 => RenL T0 (lifted_shRen E0 T E)
  end.

Fixpoint lifted_shSub u E0 T (E:Env u) : SubT (E0++E) (E0++T::E) :=
  match E0 with
    | nil => shSub T
    | T0::E0 => SubL T0 (lifted_shSub E0 T E)
  end.

Fixpoint lifted_singsub u w (sty: Ty w) : subT (S u + w) (u + w) :=
  match u with
    | 0 => [|sty|]
    | S u' => subL (lifted_singsub sty)
  end.

Fixpoint lifted_singSub u E0 E T (stv: Val u E T) : SubT (E0++T::E) (E0++E) :=
  match E0 with
    | nil => {|stv|}
    | T0::E0 => SubL T0 (lifted_singSub E0 stv)
  end.

Lemma lifted_shsub_shren: forall u w (var: varT (u+w)),
  lifted_shsub u w var = TYVAR (lifted_shren u w var).
Proof. 
induction u; first by [].
by intros; depdestr var; [ | simpl; rewrite IHu].
Qed.

Lemma lifted_shSub_shRen: forall u E0 T0 (E:Env u) T (var : VarT _ T),
  lifted_shSub E0 T0 E var = VAR (lifted_shRen E0 T0 E var).
Proof. 
induction E0; first by [].
by intros; depdestr var; [ | simpl; rewrite IHE0].
Qed.

Lemma subSub_lifted_shSub: forall u w (sub:subT u w) (E0:Env u) T0 E,
  {{ SSub,PSub # subSub sub (lifted_shSub E0 T0 E) ==
     lifted_shSub (subEnv sub E0) (subTy sub T0) (subEnv sub E) }}.
Proof. 
induction E0; intros.
- by simpl; rewrite subSub_shSub.
- simpl lifted_shSub; Hrewrite (sym_Heq (IHE0 T0 E)).
  by rewrite subSub_SubL; simpl; Helimc.
Qed.

Lemma subSub_lifted_singSub:
  forall u w (sub:subT u w) (E0:Env u) T0 E (stv: Val u E T0),
  {{ SSub,PSub # subSub sub (lifted_singSub E0 stv) ==
     lifted_singSub (subEnv sub E0) (subVal sub stv) }}.
Proof.
induction E0; intros.
- by simpl lifted_singSub; rewrite subSub_consMap; rewrite subSub_idSub.
- simpl lifted_singSub; Hrewrite (sym_Heq (IHE0 T0 E stv)).
  by rewrite subSub_SubL; simpl; Helimc.
Qed.

(****************
   Definitions
 ****************)

Inductive UTy : Type :=
  | UTYVAR : nat -> UTy
  | UUNIT : UTy
  | UINT : UTy
  | UARROW : UTy -> UTy -> UTy
  | UPROD : UTy -> UTy -> UTy
  | USUM  : UTy -> UTy -> UTy
  | UALL : UTy -> UTy
  | UEXIST : UTy -> UTy
.

Inductive UVal : Type :=
| UVAR : nat -> UVal
| UVUNIT : UVal
| UVINT  : nat -> UVal
| UPAIR  : UVal -> UVal -> UVal
| UINL   : UVal -> UVal
| UINR   : UVal -> UVal
| ULAM   : UExp -> UVal
| UREC   : UExp -> UVal
| UTLAM  : UVal -> UVal
| UPACK  : UTy -> UVal -> UVal
with UExp : Type :=
| UVAL   : UVal -> UExp
| UOPN   : OperationNat -> UVal -> UVal -> UExp
| UOPB   : OperationBool -> UVal -> UVal -> UExp
| UFST   : UTy -> UVal -> UExp
| USND   : UTy -> UVal -> UExp
| UCASE  : UVal -> UTy -> UExp -> UTy -> UExp -> UExp
| ULET   : UExp -> UTy -> UExp -> UExp
| UAPP   : UTy -> UVal -> UVal -> UExp
| UTAPP  : UTy -> UVal -> UTy -> UExp
| UUNPK  : UTy -> UVal -> UExp -> UExp
.

Definition UVBOOL (b:bool) : UVal :=
  if b then (UINL UVUNIT) else (UINR UVUNIT).

Scheme UVal_ind2 := Induction for UVal Sort Prop
  with UExp_ind2 := Induction for UExp Sort Prop.

Combined Scheme UValExp_ind from UVal_ind2, UExp_ind2.

Fixpoint UshTy i (ty: UTy) : UTy :=
  match ty with
    | UTYVAR j => UTYVAR (if blt_nat j i then j else S j)
    | UUNIT => UUNIT
    | UINT => UINT
    | UARROW ty1 ty2 => UARROW (UshTy i ty1) (UshTy i ty2)
    | UPROD ty1 ty2 => UPROD (UshTy i ty1) (UshTy i ty2)
    | USUM ty1 ty2 => USUM (UshTy i ty1) (UshTy i ty2)
    | UALL ty => UALL (UshTy (S i) ty)
    | UEXIST ty => UEXIST (UshTy (S i) ty)
  end.

Fixpoint UshTyk i k (ty: UTy) : UTy :=
  match k with 
    | 0 => ty
    | S k' => UshTy i (UshTyk i k' ty) 
  end.

Fixpoint UsubTy i (sty: UTy) (ty: UTy) : UTy :=
  match ty with
    | UTYVAR j => if blt_nat j i then UTYVAR j 
                  else (if beq_nat j i then sty
                  else UTYVAR (Peano.pred j))
    | UUNIT => UUNIT
    | UINT => UINT
    | UARROW ty1 ty2 => UARROW (UsubTy i sty ty1) (UsubTy i sty ty2)
    | UPROD ty1 ty2 => UPROD (UsubTy i sty ty1) (UsubTy i sty ty2)
    | USUM ty1 ty2 => USUM (UsubTy i sty ty1) (UsubTy i sty ty2)
    | UALL ty => UALL (UsubTy (S i) (UshTy 0 sty) ty)
    | UEXIST ty => UEXIST (UsubTy (S i) (UshTy 0 sty) ty)
  end.

Fixpoint UshVal i (tv: UVal) : UVal :=
  match tv with
    | UVAR j => UVAR j
    | UVUNIT => UVUNIT
    | UVINT n => UVINT n
    | UPAIR tv1 tv2 => UPAIR (UshVal i tv1) (UshVal i tv2)
    | UINL tv => UINL (UshVal i tv)
    | UINR tv => UINR (UshVal i tv)
    | ULAM tc => ULAM (UshExp i tc)
    | UREC tc => UREC (UshExp i tc)
    | UTLAM tv => UTLAM (UshVal (S i) tv)
    | UPACK ty tv => UPACK (UshTy i ty) (UshVal i tv)
  end
with UshExp i (tc: UExp) : UExp :=
  match tc with
    | UVAL tv => UVAL (UshVal i tv)
    | UOPN op tv1 tv2 => UOPN op (UshVal i tv1) (UshVal i tv2)
    | UOPB op tv1 tv2 => UOPB op (UshVal i tv1) (UshVal i tv2)
    | UFST ty tv => UFST (UshTy i ty) (UshVal i tv)
    | USND ty tv => USND (UshTy i ty) (UshVal i tv)
    | UCASE tv ty1 tc1 ty2 tc2 => UCASE (UshVal i tv) 
      (UshTy i ty1) (UshExp i tc1) (UshTy i ty2) (UshExp i tc2)
    | ULET tc1 ty tc2 => ULET (UshExp i tc1) (UshTy i ty) (UshExp i tc2)
    | UAPP ty tv1 tv2 => UAPP (UshTy i ty) (UshVal i tv1) (UshVal i tv2)
    | UTAPP tyr tv ty => UTAPP (UshTy (S i) tyr) (UshVal i tv) (UshTy i ty)
    | UUNPK tyr tv tc => 
      UUNPK (UshTy (S i) tyr) (UshVal i tv) (UshExp (S i) tc)
  end.

Fixpoint UShVal i (tv: UVal) : UVal :=
  match tv with
    | UVAR j => UVAR (if blt_nat j i then j else S j)
    | UVUNIT => UVUNIT
    | UVINT n => UVINT n
    | UPAIR tv1 tv2 => UPAIR (UShVal i tv1) (UShVal i tv2)
    | UINL tv => UINL (UShVal i tv)
    | UINR tv => UINR (UShVal i tv)
    | ULAM tc => ULAM (UShExp (S i) tc)
    | UREC tc => UREC (UShExp (S (S i)) tc)
    | UTLAM tv => UTLAM (UShVal i tv)
    | UPACK ty tv => UPACK ty (UShVal i tv)
  end
with UShExp i (tc: UExp) : UExp :=
  match tc with
    | UVAL tv => UVAL (UShVal i tv)
    | UOPN op tv1 tv2 => UOPN op (UShVal i tv1) (UShVal i tv2)
    | UOPB op tv1 tv2 => UOPB op (UShVal i tv1) (UShVal i tv2)
    | UFST ty tv => UFST ty (UShVal i tv)
    | USND ty tv => USND ty (UShVal i tv)
    | UCASE tv ty1 tc1 ty2 tc2 => 
      UCASE (UShVal i tv) ty1 (UShExp (S i) tc1) ty2 (UShExp (S i) tc2)
    | ULET tc1 ty tc2 => ULET (UShExp i tc1) ty (UShExp (S i) tc2)
    | UAPP ty tv1 tv2 => UAPP ty (UShVal i tv1) (UShVal i tv2)
    | UTAPP tyr tv ty => UTAPP tyr (UShVal i tv) ty
    | UUNPK tyr tv tc => UUNPK tyr (UShVal i tv) (UShExp (S i) tc)
  end.

Fixpoint UShValk i k (tv: UVal) : UVal :=
  match k with
    | 0 => tv
    | S k' => UShVal i (UShValk i k' tv)
  end.

Fixpoint UShExpk i k (tv: UExp) : UExp :=
  match k with
    | 0 => tv
    | S k' => UShExp i (UShExpk i k' tv)
  end.

Fixpoint UsubVal i (sty: UTy) (tv: UVal) : UVal :=
  match tv with
    | UVAR j => UVAR j
    | UVUNIT => UVUNIT
    | UVINT n => UVINT n
    | UPAIR tv1 tv2 => UPAIR (UsubVal i sty tv1) (UsubVal i sty tv2)
    | UINL tv => UINL (UsubVal i sty tv)
    | UINR tv => UINR (UsubVal i sty tv)
    | ULAM tc => ULAM (UsubExp i sty tc)
    | UREC tc => UREC (UsubExp i sty tc)
    | UTLAM tv => UTLAM (UsubVal (S i) (UshTy 0 sty) tv)
    | UPACK ty tv => UPACK (UsubTy i sty ty) (UsubVal i sty tv)
  end
with UsubExp i (sty: UTy) (tc: UExp) : UExp :=
  match tc with
    | UVAL tv => UVAL (UsubVal i sty tv)
    | UOPN op tv1 tv2 => UOPN op (UsubVal i sty tv1) (UsubVal i sty tv2)
    | UOPB op tv1 tv2 => UOPB op (UsubVal i sty tv1) (UsubVal i sty tv2)
    | UFST ty tv => UFST (UsubTy i sty ty) (UsubVal i sty tv)
    | USND ty tv => USND (UsubTy i sty ty) (UsubVal i sty tv)
    | UCASE tv ty1 tc1 ty2 tc2 => 
      UCASE (UsubVal i sty tv) 
        (UsubTy i sty ty1) (UsubExp i sty tc1) 
        (UsubTy i sty ty2) (UsubExp i sty tc2)
    | ULET tc1 ty tc2 => 
      ULET (UsubExp i sty tc1) (UsubTy i sty ty) (UsubExp i sty tc2)
    | UAPP ty tv1 tv2 => 
      UAPP (UsubTy i sty ty) (UsubVal i sty tv1) (UsubVal i sty tv2)
    | UTAPP tyr tv ty => 
      UTAPP (UsubTy (S i) (UshTy 0 sty) tyr) 
       (UsubVal i sty tv) (UsubTy i sty ty)
    | UUNPK tyr tv tc => 
      UUNPK (UsubTy (S i) (UshTy 0 sty) tyr) 
       (UsubVal i sty tv) (UsubExp (S i) (UshTy 0 sty) tc)
  end.

Fixpoint USubVal i (stv: UVal) (tv: UVal) : UVal :=
  match tv with
    | UVAR j => if blt_nat j i then UVAR j 
                else (if beq_nat j i then stv
                else UVAR (Peano.pred j))
    | UVUNIT => UVUNIT
    | UVINT n => UVINT n
    | UPAIR tv1 tv2 => UPAIR (USubVal i stv tv1) (USubVal i stv tv2)
    | UINL tv => UINL (USubVal i stv tv)
    | UINR tv => UINR (USubVal i stv tv)
    | ULAM tc => ULAM (USubExp (S i) (UShVal 0 stv) tc)
    | UREC tc => UREC (USubExp (S (S i)) (UShVal 0 (UShVal 0 stv)) tc)
    | UTLAM tv => UTLAM (USubVal i (UshVal 0 stv) tv)
    | UPACK ty tv => UPACK ty (USubVal i stv tv)
  end
with USubExp i (stv: UVal) (tc: UExp) : UExp :=
  match tc with
    | UVAL tv => UVAL (USubVal i stv tv)
    | UOPN op tv1 tv2 => UOPN op (USubVal i stv tv1) (USubVal i stv tv2)
    | UOPB op tv1 tv2 => UOPB op (USubVal i stv tv1) (USubVal i stv tv2)
    | UFST ty tv => UFST ty (USubVal i stv tv)
    | USND ty tv => USND ty (USubVal i stv tv)
    | UCASE tv ty1 tc1 ty2 tc2 => 
      UCASE (USubVal i stv tv) ty1 
       (USubExp (S i) (UShVal 0 stv) tc1) ty2 (USubExp (S i) (UShVal 0 stv) tc2)
    | ULET tc1 ty tc2 => 
      ULET (USubExp i stv tc1) ty (USubExp (S i) (UShVal 0 stv) tc2)
    | UAPP ty tv1 tv2 => UAPP ty (USubVal i stv tv1) (USubVal i stv tv2)
    | UTAPP tyr tv ty => UTAPP tyr (USubVal i stv tv) ty
    | UUNPK tyr tv tc => 
      UUNPK tyr (USubVal i stv tv) (USubExp (S i) (UShVal 0 (UshVal 0 stv)) tc)
  end.

Reserved Notation "e |- i => v" (at level 70, no associativity). 

(* only consider UEv for closed terms *)

Inductive UEv: nat -> UExp -> UVal -> Prop :=
| e_UVal : forall v, 
  UVAL v |- 0 => v
| e_UOpN  : forall op n1 n2,
  UOPN op (UVINT n1) (UVINT n2) |- 1 => UVINT (OpSemNat op n1 n2)
| e_UOpB  : forall op n1 n2, 
  UOPB op (UVINT n1) (UVINT n2) |- 1 => UVBOOL (OpSemBool op n1 n2)
| e_UFst  : forall ty v1 v2,
  UFST ty (UPAIR v1 v2) |- 1 => v1
| e_USnd  : forall ty v1 v2,
  USND ty (UPAIR v1 v2) |- 1 => v2
| e_UCaseL : forall k v ty1 e1 ty2 e2 v1,
  USubExp 0 v e1 |- k => v1 -> UCASE (UINL v) ty1 e1 ty2 e2 |- S k => v1
| e_UCaseR : forall k v ty1 e1 ty2 e2 v2,
  USubExp 0 v e2 |- k => v2 -> UCASE (UINR v) ty1 e1 ty2 e2 |- S k => v2
| e_UApp : forall k ty e v1 v2,
  USubExp 0 v1 e |- k => v2 -> UAPP ty (ULAM e) v1 |- S k => v2
| e_URApp : forall k ty e v1 v2,
  USubExp 0 (UREC e) (USubExp 0 v1 e) |- k => v2 -> 
  UAPP ty (UREC e) v1 |- S k => v2
| e_UTApp : forall v tyr ty,
  UTAPP tyr (UTLAM v) ty |- 1 => UsubVal 0 ty v
| e_UUnpk : forall k tyr ty v1 v2 e,
  USubExp 0 v1 (UsubExp 0 ty e) |- k => v2 -> 
  UUNPK tyr (UPACK ty v1) e |- S k => v2
| e_ULet : forall k1 k2 e1 ty e2 v1 v2,
  e1 |- k1 => v1 -> USubExp 0 v1 e2 |- k2 => v2 -> 
  ULET e1 ty e2 |- S (k1+k2) => v2
where "e |- i => v" := (UEv i e v).

Fixpoint reflectTy u (ty: Ty u) : UTy :=
  match ty with
    | TYVAR var => UTYVAR (idxvar var)
    | UNIT => UUNIT
    | INT => UINT
    | ARROW ty1 ty2 => UARROW (reflectTy ty1) (reflectTy ty2)
    | PROD ty1 ty2 => UPROD (reflectTy ty1) (reflectTy ty2)
    | SUM ty1 ty2 => USUM (reflectTy ty1) (reflectTy ty2)
    | ALL ty => UALL (reflectTy ty)
    | EXIST ty => UEXIST (reflectTy ty)
  end.
    
Fixpoint reflectVal u E T (tv: Val u E T) : UVal :=
  match tv with
    | VAR _ var => UVAR (idxVar var)
    | VUNIT => UVUNIT
    | VINT i => UVINT i
    | PAIR _ _ v1 v2 => UPAIR (reflectVal v1) (reflectVal v2)
    | INL _ _ v1 => UINL (reflectVal v1)
    | INR _ _ v2 => UINR (reflectVal v2)
    | LAM _ _ e => ULAM (reflectExp e)
    | REC _ _ e => UREC (reflectExp e)
    | TLAM _ v => UTLAM (reflectVal v)
    | PACK _ ty v => UPACK (reflectTy ty) (reflectVal v)
  end
with reflectExp u E T (tc: Exp u E T) : UExp :=
  match tc with
    | VAL _ v => UVAL (reflectVal v)
    | OPN op v1 v2 => UOPN op (reflectVal v1) (reflectVal v2)
    | OPB op v1 v2 => UOPB op (reflectVal v1) (reflectVal v2)
    | FST ty1 ty2 v => UFST (reflectTy ty2) (reflectVal v)
    | SND ty1 ty2 v => USND (reflectTy ty1) (reflectVal v)
    | CASE ty1 ty2 _ v e1 e2 => 
      UCASE (reflectVal v) (reflectTy ty1) (reflectExp e1) 
      (reflectTy ty2) (reflectExp e2)
    | LET ty1 ty2 e1 e2 => ULET (reflectExp e1) (reflectTy ty1) (reflectExp e2)
    | APP ty1 _ v1 v2 => UAPP (reflectTy ty1) (reflectVal v1) (reflectVal v2)
    | TAPP tyr f ty => UTAPP (reflectTy tyr) (reflectVal f) (reflectTy ty)
    | UNPK tyr _ v e => UUNPK (reflectTy tyr) (reflectVal v) (reflectExp e)
  end.

(****************
   Properties
 ****************)

Lemma reflectVal_VBOOL: forall u E b,
  reflectVal (@VBOOL u E b) = UVBOOL b.
Proof. by destruct b; auto. Qed.

Lemma reflectVal_cast: forall u E T E' T' (pf: (E,T) = (E',T')) (tv: Val u E T),
  reflectVal (<<SVal,PVal # pf>> tv) = reflectVal tv.
Proof. by intros; Hsimplc. Qed.

Lemma reflectExp_cast: forall u E T E' T' (pf: (E,T) = (E',T')) (tv: Exp u E T),
  reflectExp (<<SExp,PExp # pf>> tv) = reflectExp tv.
Proof. by intros; Hsimplc. Qed.

Lemma reflectTy_inj: forall u (ty ty': Ty u),
  reflectTy ty = reflectTy ty' -> ty = ty'.
Proof.
induction ty; intros; depdestr ty'; 
  try (erewrite IHty; last by Apply);
  try (erewrite IHty1; last by Apply);
  try (erewrite IHty2; last by Apply);
  auto.
- by f_equal; eapply idxvar_inj.
Qed.

Lemma reflectValExp_inj: forall u E,
  (forall T (tv tv': Val u E T), reflectVal tv = reflectVal tv' -> tv = tv') /\
  (forall T (tv tv': Exp u E T), reflectExp tv = reflectExp tv' -> tv = tv').
Proof. 
apply ValExp_ind; intros;
  try (depdestr tv'; auto;
       try (assert (Htyinj := reflectTy_inj _ _ H1); subst);
       try (assert (Htyinj := reflectTy_inj _ _ H2); subst);
       try (assert (Htyinj := reflectTy_inj _ _ H3); subst);
       try (assert (Htyinj := reflectTy_inj _ _ H5); subst);
       try (erewrite H; last by Apply); auto;
       try (erewrite H0; last by Apply); auto;
       try (erewrite H1; last by Apply); auto).
- by inversion H; erewrite (idxVar_inj _); auto.
- by rewrite -> (Hrw (sym_Heq (Heq_split SExp 
    (f_equal (fun T => (_,T)) H2) (proj2 (Heq_JMeq _ _) H3)))), 
    reflectExp_cast in H1; discriminate.
- by rewrite -> (Hrw (sym_Heq (Heq_split SExp 
    (f_equal (fun T => (_,T)) H2) (proj2 (Heq_JMeq _ _) H3)))), 
    reflectExp_cast in H1; discriminate.
- by rewrite -> (Hrw (sym_Heq (Heq_split SExp 
    (f_equal (fun T => (_,T)) H1) (proj2 (Heq_JMeq _ _) H2)))), 
    reflectExp_cast in H0; discriminate.
- by rewrite -> (Hrw (sym_Heq (Heq_split SExp 
    (f_equal (fun T => (_,T)) H1) (proj2 (Heq_JMeq _ _) H2)))), 
    reflectExp_cast in H0; discriminate.
- rewrite -> (Hrw (sym_Heq (Heq_split SExp 
    (f_equal (fun T => (_,T)) H1) (proj2 (Heq_JMeq _ _) H2)))), 
    reflectExp_cast in H0.
  simpl in H0. depdestr H0. 
  assert (Htyinj := reflectTy_inj _ _ H0); subst.
  assert (Htyinj := reflectTy_inj _ _ H4); subst.
  by rewrite <- H2, (H _ H3).
Qed.

Definition reflectVal_inj u E := proj1 (@reflectValExp_inj u E).
Definition reflectExp_inj u E := proj2 (@reflectValExp_inj u E).

Lemma UShValExp_UshValExp: 
  (forall utv i j, UShVal i (UshVal j utv) = UshVal j (UShVal i utv)) /\
  (forall utv i j, UShExp i (UshExp j utv) = UshExp j (UShExp i utv)).
Proof.
apply UValExp_ind; intros; simpl;
  try rewrite H; try rewrite H0; try rewrite H1; auto.
Qed.

Definition UShVal_UshVal := proj1 (UShValExp_UshValExp).
Definition UShExp_UshExp := proj2 (UShValExp_UshValExp).

Lemma UShValk_UshVal: forall i k j utv,
  UShValk i k (UshVal j utv) = UshVal j (UShValk i k utv).
Proof.
by induction k; intros; simpl; [ | rewrite <- (proj1 UShValExp_UshValExp), IHk].
Qed.

Lemma UShExpk_UshExp: forall i k j utv,
  UShExpk i k (UshExp j utv) = UshExp j (UShExpk i k utv).
Proof.
by induction k; intros; simpl; [ | rewrite <- (proj2 UShValExp_UshValExp), IHk].
Qed.

Lemma UsubTy_closed_general: forall u (ty: Ty u) i sty,
  i >= u -> UsubTy i sty (reflectTy ty) = reflectTy ty.
Proof.
induction ty; intros; simpl; 
  try rewrite IHty; try rewrite IHty1; try rewrite IHty2; auto with arith.
- have Hi := idxvar_range v; rewrite blt_nat_true; by [ | omega].
Qed.

Definition UsubTy_closed ty i sty := 
  @UsubTy_closed_general 0 ty i sty (le_O_n _).

Lemma UsubValExp_closed_general: forall u E,
  (forall T (tv: Val u E T) sty i, 
    i >= u -> UsubVal i sty (reflectVal tv) = reflectVal tv) /\
  (forall T (tv: Exp u E T) sty i, 
    i >= u -> UsubExp i sty (reflectExp tv) = reflectExp tv).
Proof.
apply ValExp_ind; intros; 
  simpl; try (rewrite H); try (rewrite H0); try (rewrite H1);
  simpl; try rewrite !UsubTy_closed_general; try omega; auto.
Qed.

Definition UsubVal_closed E T tv sty i :=
  proj1 (@UsubValExp_closed_general 0 E) T tv sty i (le_O_n _).
Definition UsubExp_closed E T tv sty i :=
  proj2 (@UsubValExp_closed_general 0 E) T tv sty i (le_O_n _).

Lemma USubValExp_closed_general: forall u E,
  (forall T (tv: Val u E T) stv i, 
    i >= length E -> USubVal i stv (reflectVal tv) = reflectVal tv) /\
  (forall T (tv: Exp u E T) stv i, 
    i >= length E -> USubExp i stv (reflectExp tv) = reflectExp tv).
Proof.
apply ValExp_ind; intros; 
  simpl; try (rewrite H); try (rewrite H0); try (rewrite H1);
  simpl; try (rewrite subEnv_length); try omega; auto.
- have Hidxv := idxVar_range v.
  by simpl; rewrite blt_nat_true; [ | omega].
Qed.

Definition USubVal_closed u T tv stv i :=
  proj1 (@USubValExp_closed_general u []) T tv stv i (le_O_n _).
Definition USubExp_closed u T tv stv i :=
  proj2 (@USubValExp_closed_general u []) T tv stv i (le_O_n _).

Lemma UshTy_closed_general: forall u (ty: Ty u) i,
  i >= u -> UshTy i (reflectTy ty) = reflectTy ty.
Proof.
induction ty; intros; simpl; 
  try rewrite IHty; try rewrite IHty1; try rewrite IHty2; auto with arith.
- have Hi := idxvar_range v; rewrite blt_nat_true; by [ | omega].
Qed.

Definition UshTy_closed ty i := @UshTy_closed_general 0 ty i (le_O_n _).

Lemma UshValExp_closed_general: forall u E,
  (forall T (tv: Val u E T) i, 
    i >= u -> UshVal i (reflectVal tv) = reflectVal tv) /\
  (forall T (tv: Exp u E T) i, 
    i >= u -> UshExp i (reflectExp tv) = reflectExp tv).
Proof.
apply ValExp_ind; intros;
  simpl; try (rewrite H); try (rewrite H0); try (rewrite H1);
  simpl; try rewrite !UshTy_closed_general; try omega; auto.
Qed.

Definition UshVal_closed E T tv i :=
  proj1 (@UshValExp_closed_general 0 E) T tv i (le_O_n _).
Definition UshExp_closed E T tv i :=
  proj2 (@UshValExp_closed_general 0 E) T tv i (le_O_n _).

Lemma UShValExp_closed_general: forall u E,
  (forall T (tv: Val u E T) i,
    i >= length E -> UShVal i (reflectVal tv) = reflectVal tv) /\ 
  (forall T (tv: Exp u E T) i,
    i >= length E -> UShExp i (reflectExp tv) = reflectExp tv).
Proof.
apply ValExp_ind; intros; 
  simpl; try (rewrite H); try (rewrite H0); try (rewrite H1);
  simpl; try (rewrite subEnv_length); try omega; auto.
- have Hidxv := idxVar_range v.
  by simpl; rewrite blt_nat_true; [ | omega].
Qed.

Definition UShVal_closed u T tv i :=
  proj1 (@UShValExp_closed_general u []) T tv i (le_O_n _).
Definition UShExp_closed u T tv i :=
  proj2 (@UShValExp_closed_general u []) T tv i (le_O_n _).

Lemma reflectTy_shsub_general: forall v (ty: Ty v) u w (C: v = u + w),
  reflectTy (subTy (lifted_shsub u w) (<<Ty#C>>ty)) = UshTy u (reflectTy ty).
Proof.
induction ty; intros; Hsimplc; simpl;
  try rewrite (IHty _ _ refl); try rewrite (IHty (S _) _ refl);
  try rewrite (IHty1 _ _ refl); try rewrite (IHty1 (S _) _ refl); 
  try rewrite (IHty2 _ _ refl); try rewrite (IHty2 (S _) _ refl); auto.
- rewrite lifted_shsub_shren; simpl; f_equal.
  induction u0. 
  - simpl; rewrite blt_nat_false; auto; by omega.
  - depdestr v; first by [].
    simpl; rewrite IHu0; by destruct (blt_nat (idxvar v) u0).
Qed.

Lemma reflectTy_shsub: forall u (ty: Ty u),
  reflectTy (subTy shsub ty) = UshTy 0 (reflectTy ty).
Proof. intros; apply (reflectTy_shsub_general ty 0 _ refl). Qed.

Lemma reflectValExp_shsub_general: forall u E,
  (forall T (tv: Val u E T) v w (C: u = v + w),
    reflectVal (subVal (lifted_shsub v w) (<<SSVal#SEnvTyC C>> tv)) = 
    UshVal v (reflectVal tv)) /\
  (forall T (tv: Exp u E T) v w (C: u = v + w),
    reflectExp (subExp (lifted_shsub v w) (<<SSExp#SEnvTyC C>> tv)) = 
    UshExp v (reflectExp tv)).
Proof.
apply ValExp_ind; intros; do 2 (simpl; Hsimplc); simpl; 
  try rewrite <- (H _ _ refl); try rewrite <- (H (S _) _ refl); 
  try rewrite <- (H0 _ _ refl); try rewrite <- (H0 (S _) _ refl); 
  try rewrite <- (H1 _ _ refl); try rewrite <- (H1 (S _) _ refl);
  repeat rewrite (reflectTy_shsub_general _ _ _ refl); 
  repeat rewrite (reflectTy_shsub_general _ (S _) _ refl); auto.
- by rewrite idxVar_subVar.
Qed.

Lemma reflectVal_shsub: forall u E T (tv: Val u E T),
  reflectVal (subVal shsub tv) = UshVal 0 (reflectVal tv).
Proof. intros; apply (proj1 (reflectValExp_shsub_general _) _ tv 0 _ refl). Qed.

Lemma reflectExp_shsub: forall u E T (tv: Exp u E T),
  reflectExp (subExp shsub tv) = UshExp 0 (reflectExp tv).
Proof. intros; apply (proj2 (reflectValExp_shsub_general _) _ tv 0 _ refl). Qed.

Lemma reflectValExp_shSub_general: forall u E,
 (forall T (tv: Val u E T) E0 T0 E1 (C: E = E0 ++ E1),
  reflectVal (SubVal (lifted_shSub E0 T0 E1) (<<SVal#f_equal(fun E=>(E,T))C>>tv))
  = UShVal (length E0) (reflectVal tv)) /\
 (forall T (tv: Exp u E T) E0 T0 E1 (C: E = E0 ++ E1),
  reflectExp (SubExp (lifted_shSub E0 T0 E1) (<<SExp#f_equal(fun E=>(E,T))C>>tv))
  = UShExp (length E0) (reflectExp tv)).
Proof.
apply ValExp_ind; intros; simpl; Hsimplc; unfold_SubVE;
  try rewrite (H _ _ _ refl); try rewrite (H (_::_) _ _ refl); 
    try rewrite (H (_::_::_) _ _ refl);
  try rewrite (H0 _ _ _ refl); try rewrite (H0 (_::_) _ _ refl);
  try rewrite (H1 _ _ _ refl); try rewrite (H1 (_::_) _ _ refl);
  auto.
- rewrite lifted_shSub_shRen; simpl; f_equal.
  induction E0. 
  - simpl; rewrite blt_nat_false; auto; by omega.
  - depdestr v; first by [].
    simpl; rewrite IHE0; by destruct (blt_nat (idxVar v) (length E0)).
- specialize (H (subEnv shsub E0) (subTy shsub T0) (subEnv shsub E1) 
   (subEnv_app _ _ _)).
  rewrite subEnv_length in H; rewrite <- H.
  by Hrewritec (sym_Heq (subSub_lifted_shSub shsub E0 T0 E1)); Helimc.
- specialize (H0 (ty1::subEnv shsub E0) (subTy shsub T0) (subEnv shsub E1)
   (f_equal (fun E => ty1::E) (subEnv_app _ _ _))).
  simpl in H0; rewrite subEnv_length in H0; rewrite <- H0.
  by Hrewritec (sym_Heq (subSub_lifted_shSub shsub E0 T0 E1)); Helimc.  
Qed.

Lemma reflectVal_shSub: forall u E T (tv: Val u E T) T',
  reflectVal (SubVal (shSub T') tv) = UShVal 0 (reflectVal tv).
Proof. intros; apply (proj1 (reflectValExp_shSub_general _) _ tv nil _ _ refl). Qed.

Lemma reflectExp_shSub: forall u E T (tv: Exp u E T) T',
  reflectExp (SubExp (shSub T') tv) = UShExp 0 (reflectExp tv).
Proof. intros; apply (proj2 (reflectValExp_shSub_general _) _ tv nil _ _ refl). Qed.

Lemma reflectTy_singsub_general: forall v (ty: Ty v) u w sty (C: v = S u + w),
  reflectTy (subTy (lifted_singsub sty) (<<Ty#C>>ty)) = 
  UsubTy u (UshTyk 0 u (reflectTy sty)) (reflectTy ty).
Proof.
induction ty; intros; Hsimplc; simpl;
  try rewrite (IHty _ _ _ refl); try rewrite (IHty (S _) _ _ refl);
  try rewrite (IHty1 _ _ _ refl); try rewrite (IHty1 (S _) _ _ refl); 
  try rewrite (IHty2 _ _ _ refl); try rewrite (IHty2 (S _) _ _ refl); auto.
- induction u0; simpl in *; depdestr v; auto. 
  simpl; rewrite -> renTy_def, reflectTy_shsub, IHu0.
  destruct (lt_eq_lt_dec (idxvar v) u0) as [[?|?]|?].
  - by rewrite blt_nat_true; [simpl; rewrite blt_nat_false; [auto|omega] | auto].
  - by rewrite blt_nat_false; [rewrite beq_nat_true | omega ].
  - simpl; rewrite blt_nat_false; last by omega.
    simpl; rewrite beq_nat_false; last by omega.
    simpl; rewrite blt_nat_false; last by omega. 
    by depdestr v; [inversion l | ].
Qed.

Lemma reflectTy_singsub: forall u (sty: Ty u) ty,
  reflectTy (subTy [|sty|] ty) = UsubTy 0 (reflectTy sty) (reflectTy ty).
Proof. intros; apply (reflectTy_singsub_general (u:=0) _ sty refl). Qed.

Lemma reflectValExp_singsub_general: forall u E,
  (forall T (tv: Val u E T) v w sty (C: u = S v + w),
    reflectVal (subVal (lifted_singsub sty) (<<SSVal#SEnvTyC C>> tv)) = 
    UsubVal v (UshTyk 0 v (reflectTy sty)) (reflectVal tv)) /\
  (forall T (tv: Exp u E T) v w sty (C: u = S v + w),
    reflectExp (subExp (lifted_singsub sty) (<<SSExp#SEnvTyC C>> tv)) = 
    UsubExp v (UshTyk 0 v (reflectTy sty)) (reflectExp tv)).
Proof.
apply ValExp_ind; intros; do 2 (simpl; Hsimplc); simpl;
  try rewrite (H _ _ _ refl); try rewrite (H (S _) _ _ refl); 
  try rewrite (H0 _ _ _ refl); try rewrite (H0 (S _) _ _ refl); 
  try rewrite (H1 _ _ _ refl); try rewrite (H1 (S _) _ _ refl);
  repeat rewrite (reflectTy_singsub_general _ _ refl);
  repeat rewrite (reflectTy_singsub_general (u:=S _) _ _ refl); auto.
- by rewrite idxVar_subVar.
Qed.

Lemma reflectVal_singsub: forall u (sty:Ty u) E T (tv: Val _ E T),
  reflectVal (subVal [|sty|] tv) = UsubVal 0 (reflectTy sty) (reflectVal tv).
Proof.
by intros; apply (proj1 (reflectValExp_singsub_general _) _ tv 0 _ sty refl).
Qed.

Lemma reflectExp_singsub: forall u (sty:Ty u) E T (tv: Exp _ E T),
  reflectExp (subExp [|sty|] tv) = UsubExp 0 (reflectTy sty) (reflectExp tv).
Proof.
by intros; apply (proj2 (reflectValExp_singsub_general _) _ tv 0 _ sty refl).
Qed.

Lemma reflectVal_subL: forall u E T (sty:Ty u) (tv: Val _ E T),
  reflectVal (subVal (subL [|sty|]) tv) = 
  UsubVal 1 (UshTy 0 (reflectTy sty)) (reflectVal tv).
Proof.
intros; apply (proj1 (reflectValExp_singsub_general _) _ tv 1 _ sty refl).
Qed.

Lemma reflectExp_subL: forall u E T (sty:Ty u) (tv: Exp _ E T),
  reflectExp (subExp (subL [|sty|]) tv) = 
  UsubExp 1 (UshTy 0 (reflectTy sty)) (reflectExp tv).
Proof.
intros; apply (proj2 (reflectValExp_singsub_general _) _ tv 1 _ sty refl).
Qed.

Lemma reflectValExp_singSub_general: forall u E,
(forall T (tv: Val u E T) E0 T0 E1 (stv: Val u E1 T0) (C: E = E0 ++ T0 :: E1),
 reflectVal (SubVal (lifted_singSub E0 stv) (<<SVal#f_equal(fun E=>(E,T))C>>tv))
 = USubVal (length E0) (UShValk 0 (length E0) (reflectVal stv)) (reflectVal tv))
/\
(forall T (tv: Exp u E T) E0 T0 E1 (stv: Val u E1 T0) (C: E = E0 ++ T0 :: E1),
 reflectExp (SubExp (lifted_singSub E0 stv) (<<SExp#f_equal(fun E=>(E,T))C>>tv))
 = USubExp (length E0) (UShValk 0 (length E0) (reflectVal stv)) (reflectExp tv)).
Proof.
apply ValExp_ind; intros; simpl; Hsimplc; unfold_SubVE; simpl; 
  try rewrite (H _ _ _ _ refl); try rewrite (H (_::_) _ _ _ refl); 
  try rewrite (H (_::_::_) _ _ _ refl);
  try rewrite (H0 _ _ _ _ refl); try rewrite (H0 (_::_) _ _ _ refl);
  try rewrite (H1 _ _ _ _ refl); try rewrite (H1 (_::_) _ _ _ refl);
  auto.
- induction E0.
  - by simpl; rewrite blt_nat_false; [depdestr v | omega].
  - depdestr v; first by [].
    simpl in *; rewrite -> RenVal_def, reflectVal_shSub, IHE0; last by [].
    remember (blt_nat (idxVar v) (length E0)) as n; destruct n.
    - by simpl; rewrite blt_nat_false; [ | omega].
    - remember (beq_nat (idxVar v) (length E0)) as n; destruct n; first by [].
      simpl; rewrite blt_nat_false; last by omega.
      destruct (idxVar v); last by [].
      destruct (length E0); [discriminate Heqn0 | discriminate Heqn].
- specialize (H (subEnv shsub E0) (subTy shsub T0) (subEnv shsub E1)
   (subVal shsub stv) (subEnv_app _ _ _)).
  rewrite -> subEnv_length, reflectVal_shsub, UShValk_UshVal in H.
  rewrite <- H.
  by Hrewritec (sym_Heq (subSub_lifted_singSub shsub E0 stv)); Helimc.
- specialize (H0 (ty1::subEnv shsub E0) (subTy shsub T0) (subEnv shsub E1)
   (subVal shsub stv) (f_equal (fun E => ty1::E) (subEnv_app _ _ _))).
  simpl in H0; rewrite -> subEnv_length, reflectVal_shsub, UShValk_UshVal in H0.
  rewrite <- H0.
  by Hrewritec (sym_Heq (subSub_lifted_singSub shsub E0 stv)); Helimc.
Qed.

Lemma reflectVal_singSub: forall u E T T' (stv: Val u E T') (tv: Val u _ T),
  reflectVal (SubVal {|stv|} tv) = USubVal 0 (reflectVal stv) (reflectVal tv).
Proof. 
intros; apply (proj1 (reflectValExp_singSub_general _) _ tv [] _ _ stv refl).
Qed.

Lemma reflectExp_singSub: forall u E T T' (stv: Val u E T') (tv: Exp u _ T),
  reflectExp (SubExp {|stv|} tv) = USubExp 0 (reflectVal stv) (reflectExp tv).
Proof. 
intros; apply (proj2 (reflectValExp_singSub_general _) _ tv [] _ _ stv refl).
Qed.

(* 
   Correctness of reflection 
*)

Lemma UEv_reflects_Ev_general: forall ty (tc: CExp ty) k utv,
  reflectExp tc |- k => utv ->
  (exists tv : CVal ty, reflectVal tv = utv) /\
  (forall tv, reflectVal tv = utv -> tc |= k => tv).
Proof.
move => ty tc k utv.
remember (reflectExp tc) as utc.
move => HUEv; move : HUEv ty tc Hequtc; elim; intros; subst.
- depdestr tc; split; first by eexists.
  by move => tv Htv; inversion Htv; erewrite reflectVal_inj.
- depdestr tc, v0, v; split; first by exists (@VINT 0 [] (OpSemNat o n0 n)).
  by move => tv Htv; depdestr tv.
- depdestr tc, v0, v; split.
  - by exists (@VBOOL 0 [] (OpSemBool o n0 n)); apply reflectVal_VBOOL.
  - move => tv Htv.
    depdestr tv; remember (OpSemBool o n0 n) as b; destruct b; inversion H.
    depdestr tv; eapply eq_ind; first (by constructor). by rewrite <- Heqb.
    depdestr tv; eapply eq_ind; first (by constructor). by rewrite <- Heqb.
- by depdestr tc, v; split; [eexists | intros; erewrite reflectVal_inj].
- by depdestr tc, v; split; [eexists | intros; erewrite reflectVal_inj].
- depdestr tc, v0.
  specialize (H0 _ (SubExp {|v0|} tc1) (sym_eq (reflectExp_singSub _ _))).
  destruct H0 as [[tv Hex] Hall].
  split; by [eexists; Apply | intros; constructor; Apply].
- depdestr tc, v0.
  specialize (H0 _ (SubExp {|v0|} tc2) (sym_eq (reflectExp_singSub _ _))).
  destruct H0 as [[tv Hex] Hall].
  split; by [eexists; Apply | intros; constructor; Apply].
- depdestr tc, v.
  specialize (H0 _ (SubExp {|v0|} e0) (sym_eq (reflectExp_singSub _ _))).
  destruct H0 as [[tv Hex] Hall]; subst.
  split; by [eexists | intros; constructor; Apply].
- depdestr tc, v.
  specialize (H0 _ (SubExp {|v0, REC e0|} e0)).
  replace {|v0,REC e0|} with ({|REC e0|} @SS@ {|SubVal (shSub _) v0|}) in H0
   by simpl_Sub.
  rewrite <-SubExp_SS,!reflectExp_singSub, reflectVal_shSub, UShVal_closed in H0.
  specialize (H0 refl); destruct H0 as [[tv Hex] Hall]; subst.
  by split; [eexists | move :Hall; simpl_Sub].
- depdestr tc, v0; split.
  - by exists (subVal [|ty'|] v0); apply (reflectVal_singsub _ v0).
  - intros; eapply eq_ind; first by constructor. 
    erewrite reflectVal_inj; by [ | rewrite (reflectVal_singsub _ v0)].
- depdestr tc, v.
  specialize (H0 _ (SubExp {|v|} (subExp [|ty'|] tc))).
  rewrite reflectExp_singSub in H0; rewrite (reflectExp_singsub _ tc) in H0.
  specialize (H0 refl); destruct H0 as [[tv Hex] Hall]; subst.
  split. 
  - have C:([]:Env 0,subTy [|ty'|] (subTy shsub ty2)) = ([],ty2) by simpl_sub.
    by exists (<<SVal,PVal#C>> tv); rewrite reflectVal_cast.
  - by intros; constructor; Apply; rewrite reflectVal_cast.
- depdestr tc. 
  specialize (H0 _ tc1 refl); destruct H0 as [[tv1 Hex1] Hall1].
  rewrite <- Hex1 in H2.
  specialize (H2 _ (SubExp {|tv1|} tc2) (sym_eq (reflectExp_singSub tv1 tc2))).
  destruct H2 as [[tv2 Hex2] Hall2]; subst.
  by split; [eexists | intros; econstructor; [eapply Hall1 | auto]].
Qed.

Lemma UEv_reflects_Ev: forall ty (tc: CExp ty) k (tv : CVal ty),
  reflectExp tc |- k => reflectVal tv -> tc |= k => tv.
Proof. 
by move => ty tc k tv Htc; apply (proj2 (UEv_reflects_Ev_general _ Htc)).
Qed.

(*
   Automation
*)
Ltac reflect_simpl := (
  simpl; rewrite -> ?reflectVal_cast, ?reflectExp_cast;
  simpl; rewrite -> ?reflectVal_shSub, ?UShVal_closed, ?USubVal_closed;
  simpl; 
  repeat (match goal with [ |- context[ reflectVal (subVal shsub ?tv)] ]
            => rewrite (reflectVal_shsub tv)
          end);
  repeat (match goal with [ |- context[ reflectVal (subVal (subL [|?T|]) ?tv)] ]
            => rewrite (reflectVal_subL T tv)
          end);
  repeat (match goal with [ |- context[ reflectVal (subVal [|?T|] ?tv)] ]
            => rewrite (reflectVal_singsub T tv)
          end);
  rewrite -> ?UshVal_closed, ?UsubVal_closed;
  simpl; rewrite -> ?renTy_def, ?reflectTy_shsub, ?UshTy_closed, ?UsubTy_closed).
