(*
  Utility stuff for
   Realizability and Compositional Compiler Correctness 
      for a Polymorphic Language
   N Benton & C-K Hur
   April 2010
*)

Require Import List Omega Program Heq.

Set Implicit Arguments.

Tactic Notation "app" tactic(tac) "in" hyp(H) := 
  (Hhide_concl; revert H; tac; intros; Hshow_concl).

Ltac args2 tac H1 H2 := (tac H1; tac H2).
Ltac args3 tac H1 H2 H3 := (tac H1; tac H2; tac H3).
Ltac args4 tac H1 H2 H3 H4 := (tac H1; tac H2; tac H3; tac H4).
Ltac args5 tac H1 H2 H3 H4 H5 := (tac H1; tac H2; tac H3; tac H4; tac H5).

Ltac extension H := extensionality H.

Tactic Notation "depdestr" ident(H) := (dependent destruction H).
Tactic Notation "depdestr" ident(H1) "," ident(H2) :=
  (depdestr H1; depdestr H2).
Tactic Notation "depdestr" ident(H1) "," ident(H2) "," ident(H3) :=
  (depdestr H1; depdestr H2; depdestr H3).
Tactic Notation "depdestr" ident(H1) "," ident(H2) "," ident(H3) "," ident(H4) :=
  (depdestr H1; depdestr H2; depdestr H3; depdestr H4).

Ltac eabs tm :=
  let ty := type of tm in
  let temp := fresh "__temp__" in
    evar (temp : ty); replace tm with temp; unfold temp; clear temp.

Ltac eprop := let ep := fresh "eprop" in let Hep := fresh "Hep" in
  evar (ep : Prop); assert (Hep : ep); try revert Hep; unfold ep; clear ep.

Ltac Apply := match goal with [ H : _ |- _ ] => apply H end.

Ltac EApply := match goal with [ H : _ |- _ ] => eapply H end.

Declare ML Module "ssreflect".
Require Import ssreflect.

Lemma generalize_induction : forall (P:nat->Prop) gk, 
  (forall k, k <= gk -> P k) -> P gk.
Proof. by intuition. Qed.

Ltac generalize_induction gk := (pattern gk; apply generalize_induction).

Lemma generalize_conj: forall (P Q:Prop), (P /\ (P -> Q)) -> P /\ Q.
Proof. by intuition. Qed.

Fixpoint ble_nat (n m: nat) {struct n} : bool :=
  match n with
    | 0 => true
    | S n1 => match m with
                | 0 => false
                | S m1 => ble_nat n1 m1
              end
  end.

Fixpoint blt_nat (n m : nat) {struct n} : bool :=
  match m with 
    | 0 => false
    | S ml => match n with 
                | 0 => true
                | S nl => blt_nat nl ml
              end
  end.

Lemma blt_nat_true: forall m n,
  n < m -> blt_nat n m = true.
Proof. 
by induction m; move => n Hn; [inversion Hn | destruct n; simpl; intuition].
Qed.

Lemma blt_nat_false: forall m n,
  n >= m -> blt_nat n m = false.
Proof. 
induction m; move => n Hn; destruct n; simpl; auto; [inversion Hn | intuition].
Qed.

Lemma beq_nat_true: forall m n,
  n = m -> beq_nat n m = true.
Proof. 
by induction m; move => n Hn; [inversion Hn | destruct n; simpl; intuition].
Qed.

Lemma beq_nat_false: forall m n,
  n <> m -> beq_nat n m = false.
Proof. 
induction m; move => n Hn; destruct n; 
 by [contradiction Hn | | | simpl; Apply; omega].
Qed.

Lemma nth_error_val: forall (A:Type) (l: list A) n,
  n < length l -> exists a, nth_error l n = value a.
Proof.
move => A; elim => [n Hn | a l IHl]; first by inversion Hn.
case => /= [Hn | n Hn]; first by exists a.
case : (IHl n); first by omega.
move => x Hlnx; eexists; by apply Hlnx.
Qed.

Lemma nth_error_prefix: forall (A:Type) l l' n (x:A),
  nth_error l n = value x -> nth_error (l++l') n = value x.
Proof.
move => A; elim => [l' | a l IHl l']; case => //=; apply: IHl.
Qed.

Lemma nth_error_eq_nth: forall A (a:A) (l l': list A) n,
  n < length l -> nth_error (l++l') n = Some (nth n l a).
Proof.
move => A a0; elim => [ | a l Hl] l' n Hn.
- by inversion Hn.
- destruct n; first by [].
  by simpl in *; apply Hl; omega.
Qed.

Lemma cons_app: forall (A:Type) (a:A) (l:list A), a :: l = (a::nil) ++ l.
Proof. auto. Qed.
